﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Reflection;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Text.RegularExpressions;


///
/// Author: Dennis Lang - 2009
/// http://home.comcast.net/~lang.dennis/
/// 
/// Desc:  Simple X-3d model viewer
/// 
/// Features:
///   Supports loading and saving of 3D X-files.
///   Uses both on-screen and off-screen D3d rendering
///   Uses delagets to control various D3d rendering pipeline
///   Built-in HSL/RGB color selector
///   Built-in D3d render settings browse/set
///   Save/Restore settings to/from registry
/// 

// TODO
//  Add buttons to reverse face 'triangle' mesh'
//  Add buttons to generate texture u/v coordinates using x/y, x/z or y/z planes to populate u/v
//  Add tab-page to show material images.
//  Add option to draw grid mesh on specific plane, with lines labeled.
//  Add option to normalize mesh, adjust center, scale
//  Add option to optimize mesh, smooth mesh ?, etc


// Menu layout:
//
// File
//      Load
//      Save
//      Export   - mesh as csv
//      Import
//      SnapShot - image
//
// Nav
//      center
//      normalize scale -1 to 1
// View
//      cullnone
//      cull clockwise
//      cull counter clockwise
//
//      Wire
//      Solid
//  
//      Background - blue grid
//                 - solid color
//                 - image
//
//      Draw axis
//      Draw bounding box
//      Draw center 'pivot' point
//
// Normal
//      Generate
//      Remove
//      Reverse
//
// Help
//      Key bindings
//      About
//      Registry
//      conv3ds.exe config
//

namespace M3dNamespace
{
    public partial class Main : Form
    {
        // Direct 3D objects
        D3d d3d;

        D3dXfile d3dXfile;
        D3dImage d3dImage;
        D3dShape d3dCenter;
        D3dLineList d3dBoundingLines;
        D3dLineList d3dNormalLines;
        D3dLineList d3dMarkerLines;
        D3dColorLineList d3dAxis;
        D3DPanel d3dPanel;

        // Camera and xfile
        float roll  = 0.0f;
        float pitch = 0.0f;
        float yaw   = 0.0f;
        string modelName = string.Empty;
        bool lightDir = true;
        bool detailChangedEnabled = true;
        bool zoomChangeEnabled = true;

        // Save/restore settings to/from registry
        private Microsoft.Win32.RegistryKey regKey;
        private string skinStyle;


        #region ===Main routines - main, init, load and draw
        public Main(string[] cmdLineArgs)
        {
            InitializeComponent();
            Version version = Assembly.GetExecutingAssembly().GetName().Version;
            this.Text += " v" + version.Major.ToString() + "." + version.Minor.ToString();
            
            blueToolStripMenuItem_Click(null, EventArgs.Empty);
            InitViewer(cmdLineArgs);

            string appName = Application.ProductName;
            this.regKey =
                Microsoft.Win32.Registry.CurrentUser.CreateSubKey(@"Software\" + appName);
            this.LoadFromRegistry();

            string modelFn = (cmdLineArgs.Count() != 0) ? cmdLineArgs[0] : string.Empty;
            if (modelFn.Length != 0)
                LoadModel(modelFn);
        }

        #region ===== Registry

        string keyModelName = "ModelName";
        string keyCamera = "Camera";
        string keyZoom = "Zoom";
        string keyFlag = "Flag";
        string keyLocSize = "LocSize";
        string keyBookmark = "bookmark";
        string keyRenderState = "renderState";

        string keyLook = "look";
        string keySkin = "skin";
        string keyBgImage = "bgImage";

        string keyMostRecent = "mostRecent";
        
        private void LoadFromRegistry()
        {
            // Load most recent list before loading any models.
            LoadMostRecent();

            // Load previous model and settings.
            string[] strValues = new string[5] {"","","","",""};

            strValues[0] = this.regKey.GetValue(keyModelName, strValues[0]) as string;
            strValues[1] = this.regKey.GetValue(keyCamera, strValues[1]) as string;
            strValues[2] = this.regKey.GetValue(keyZoom, strValues[2]) as string;
            strValues[3] = this.regKey.GetValue(keyFlag, strValues[3]) as string;
            strValues[4] = this.regKey.GetValue(keyLocSize, strValues[4]) as string;

            // If model already loaded, keep it.
            if (this.modelName.Length != 0)
                strValues[0] = this.modelName;

            this.FromStrings(strValues);

            Microsoft.Win32.RegistryKey bookmarkKey = this.regKey.CreateSubKey(keyBookmark);
            LoadListFromRegistry(this.bookmarkView, bookmarkKey, ";");

            // Load style/skin
            Microsoft.Win32.RegistryKey lookKey = this.regKey.CreateSubKey(keyLook);
            string skinStr = lookKey.GetValue(keySkin) as string;

            switch (skinStr)
            {
                case "blue":
                    blueToolStripMenuItem_Click(null, EventArgs.Empty);
                    break;
                case "black":
                    blackToolStripMenuItem_Click(null, EventArgs.Empty);
                    break;
                case "white":
                    whiteToolStripMenuItem_Click(null, EventArgs.Empty);
                    break;
            }

            // TODO - set background image
            string bgImageStr = lookKey.GetValue(keyBgImage) as string;
        }

        private void LoadMostRecent()
        {
            /// Load most recent history
            Microsoft.Win32.RegistryKey mostRecentKey = this.regKey.CreateSubKey(keyMostRecent);
            this.recentFilesToolStripMenuItem.DropDownItems.Clear();
            foreach (string k in mostRecentKey.GetValueNames())
            {
                string fn = mostRecentKey.GetValue(k).ToString();
                this.recentFilesToolStripMenuItem.DropDownItems.Add(fn).Click += new EventHandler(mru_Click);
            }
        }

        private void AddMostRecent(string fn)
        {
            bool found = false;
            int cnt = this.recentFilesToolStripMenuItem.DropDownItems.Count;
            int idx;
            for (idx = 0; idx < cnt && !found; idx++)
                found = (this.recentFilesToolStripMenuItem.DropDownItems[idx].Text == fn);

            if (!found)
            {
                if (cnt > 5)
                    this.recentFilesToolStripMenuItem.DropDownItems.RemoveAt(0);

                this.recentFilesToolStripMenuItem.DropDownItems.Add(fn).Click += new EventHandler(mru_Click);
                Microsoft.Win32.RegistryKey mostRecentKey = this.regKey.CreateSubKey(keyMostRecent);
                
                cnt = this.recentFilesToolStripMenuItem.DropDownItems.Count;
                for (idx = 0; idx < cnt; idx++)
                {
                    string mruFn = this.recentFilesToolStripMenuItem.DropDownItems[idx].Text;
                    mostRecentKey.SetValue(idx.ToString("00"), mruFn);
                }
            }
        }

        private void mru_Click(object obj, EventArgs e)
        {
            string fn = obj.ToString();
            this.LoadModel(fn);
        }

        private void SaveToRegistry()
        {
            string[] strValues = ToStrings();

            this.regKey.SetValue(keyModelName, strValues[0]);
            this.regKey.SetValue(keyCamera, strValues[1]);
            this.regKey.SetValue(keyZoom, strValues[2]);
            this.regKey.SetValue(keyFlag, strValues[3]);
            this.regKey.SetValue(keyLocSize, strValues[4]);

            Microsoft.Win32.RegistryKey bookmarkKey = this.regKey.CreateSubKey(keyBookmark);
            SaveListToRegistry(this.bookmarkView, bookmarkKey, ";");
            Microsoft.Win32.RegistryKey renderStateKey = this.regKey.CreateSubKey(keyRenderState);
            SaveListToRegistry(this.d3dView, renderStateKey, ";");

            // Save style/skin
            Microsoft.Win32.RegistryKey lookKey = this.regKey.CreateSubKey(keyLook);
            lookKey.SetValue(keySkin, this.skinStyle);
        }

        private void SaveListToRegistry(ListView lv, Microsoft.Win32.RegistryKey key, string sep)
        {
            foreach (ListViewItem item in lv.Items)
            {
                string str = "";
                foreach (ListViewItem.ListViewSubItem subItem in item.SubItems)
                    str += ((str.Length != 0) ? sep : "") + subItem.Text;

                key.SetValue(item.Index.ToString("00"), str);
            }
        }

        private void LoadListFromRegistry(ListView lv, Microsoft.Win32.RegistryKey key, string sep)
        {
            lv.Items.Clear();
            for (int i = 0; ; i++)
            {
                string str = key.GetValue(i.ToString()) as string;
                if (str == null)
                    break;

                string[] seps = new String[1] { sep };
                string[] values = str.Split(seps, StringSplitOptions.None);
                if (values.Length == lv.Columns.Count)
                {
                    ListViewItem item = lv.Items.Add(values[0]);
                    for (int subIdx = 1; subIdx < values.Length; subIdx++)
                    {
                        item.SubItems.Add(values[subIdx]);
                    }
                }
            }
        }

        string keyLight = "Light";
        string keyDrawNormal = "DrawNormal";
        string keyDrawAxis = "DrawAxis";
        string keyDrawBbox = "DrawBbox";
        string keySingleView = "SingleView";


        /// <summary>
        /// Save settings to string array
        /// 0:modelName, 1:camera, 2:zoom, 3:flags, 4:locSize  
        /// </summary>
        string[] ToStrings()
        {
            string cameraStr = this.rollBox.Text +
                "," + this.yawBox.Text +
                "," + this.pitchBox.Text;

            string zoomStr = this.zoomXbox.Text +
                "," + this.zoomYbox.Text +
                "," + this.zoomZbox.Text;

            string flagStr = "";
            if (this.lightCkBox.Checked)
                flagStr += "," + keyLight;
            if (this.drawNormalsCkBox.Checked)
                flagStr += "," + keyDrawNormal;
            if (this.axisToolStripMenuItem.Checked)
                flagStr += "," + keyDrawAxis;
            if (this.boundingBoxToolStripMenuItem.Checked)
                flagStr += "," + keyDrawBbox;
            if (this.singleViewerToolStripMenuItem.Checked)
                flagStr += "," + keySingleView;

            string locSizeStr =   
                this.ClientRectangle.X.ToString() + "," +
                this.ClientRectangle.Y.ToString() + "," +
                this.ClientRectangle.Width.ToString() + "," +
                this.ClientRectangle.Height.ToString();

            string[] strValues = new string[5];
            strValues[0] = this.modelName;
            strValues[1] = cameraStr;
            strValues[2] = zoomStr;
            strValues[3] = flagStr;
            strValues[4] = locSizeStr;

            return strValues;
        }


        /// <summary>
        /// Parse specifications from string array.
        /// 0:modelName, 1:camera, 2:zoom, 3:flags, 4:locSize
        /// </summary>
        void FromStrings(string[] strValues)
        {
            if (this.modelName != strValues[0])
                LoadModel(strValues[0]);

            string[] camera = strValues[1].Split(',');
            if (camera.Length == 3)
            {
                this.rollBox.Text = camera[0];
                this.yawBox.Text = camera[1];
                this.pitchBox.Text = camera[2];
                CameraRYP(DtoR(ParseToF(rollBox.Text)), DtoR(ParseToF(yawBox.Text)), DtoR(ParseToF(pitchBox.Text)));
            }

            string[] zoom = strValues[2].Split(',');
            if (zoom.Length == 3)
            {
                this.zoomChangeEnabled = false;
                this.zoomXbox.Text = zoom[0];
                this.zoomYbox.Text = zoom[1];
                this.zoomZbox.Text = zoom[2];
                this.zoomChangeEnabled = true;
                this.UpdateZoom();
            }

            string flag = strValues[3];
            this.lightCkBox.Checked = flag.Contains(keyLight);
            this.drawNormalsCkBox.Checked = flag.Contains(keyDrawNormal);
            this.axisToolStripMenuItem.Checked = flag.Contains(keyDrawAxis);
            this.boundingBoxToolStripMenuItem.Checked = flag.Contains(keyDrawBbox);

            this.UpdateNormals();
            this.UpdateAxis();
            this.UpdateBounding();

            if (flag.Contains(keySingleView))
                singleViewerToolStripMenuItem_Click(null, EventArgs.Empty);
            else
                threeViewersToolStripMenuItem_Click(null, EventArgs.Empty);

            if (strValues.Length > 4)
            {
                string[] locSize = strValues[4].Split(',');
                if (locSize.Length == 4)
                {
                    int[] i4 = Array.ConvertAll<string, int>(locSize, delegate(string s) { return int.Parse(s); });
                    Rectangle rect = new Rectangle(i4[0], i4[1], i4[2], i4[3]);
                    this.Location = rect.Location;
                    this.Size = rect.Size;
                }
            }
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            this.SaveToRegistry();
            base.OnClosing(e);
        }
        #endregion

        #region ===== Style/skin
        private void UpdateStyle(Color foreColor, Color backColor, Image bgImage)
        {
            this.BackColor = backColor;

            foreach (ToolStripItem toolItem in this.mainMenu.Items)
            {
                toolItem.ForeColor = foreColor;

                ToolStripMenuItem mItem = toolItem as ToolStripMenuItem;
                foreach (ToolStripItem item2 in mItem.DropDownItems)
                {
                    item2.BackgroundImage = bgImage;
                    item2.BackColor = backColor;
                    item2.BackgroundImageLayout = System.Windows.Forms.ImageLayout.Stretch;
                    item2.ForeColor = foreColor;
                }
            }

            this.renderPage.BackColor = backColor;
            this.toolBarPage.BackColor = backColor;
            // this.lightPage.BackColor = backColor;

            this.Refresh();
        }

        private void blueToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.skinStyle = "blue";
            UpdateStyle(Color.White, Color.Blue, global::M3dNamespace.Properties.Resources.blueBar);
        }

        private void whiteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.skinStyle = "white";
            UpdateStyle(Color.Black, Color.White, null);
        }

        private void blackToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.skinStyle = "black";
            Color bgColor = Color.FromArgb(255, 30, 30, 30);
            UpdateStyle(Color.White, bgColor, global::M3dNamespace.Properties.Resources.blackBar);
        }

        #endregion

        #region ===== Drawing routines

        #region ===== When to allow drawing
        bool doDraw = true;
        Stack<bool> doDrawS = new Stack<bool>();

        private void OffDraw()
        {
            doDrawS.Push(doDraw);
            doDraw = false;
        }

        private void OnDraw()
        {
            doDraw = doDrawS.Pop();
        }
        #endregion

        public void InitViewer(string[] cmdLineArgs)
        {
#if true
            // Draw D3d directly to window
            this.d3dPanel = new D3DPanel();
            this.d3dPanel.Dock = System.Windows.Forms.DockStyle.Fill;
            this.d3dPanel.Location = new System.Drawing.Point(3, 168);
            this.d3dPanel.MouseMove += new System.Windows.Forms.MouseEventHandler(this.pictureBox_MouseMove);

            this.ulTable.Controls.Clear();
            this.ulTable.Controls.Add(this.d3dPanel, 0, 1);
            // this.ulTable.Controls.Add(this.pictureBox0, 0, 1);
            this.ulTable.Controls.Add(this.btnTable0, 0, 0);

            this.pictureBox0.Visible = false;
            this.d3d = new D3d(this.d3dPanel);

#else
            // Draw D3d using off-screen
            this.d3d = new D3d(this);
#endif
            this.d3dImage = new D3dImage(d3d);
            this.d3dBoundingLines = new D3dLineList();
            this.d3dBoundingLines.color = Color.Coral;
            this.d3dNormalLines = new D3dLineList();
            this.d3dNormalLines.color = Color.Red;
            this.d3dMarkerLines = new D3dLineList();
            this.d3dMarkerLines.color = Color.Green;
            this.d3dAxis = new D3dColorLineList();
            this.d3dCenter = new D3dShape();
            this.d3dCenter.color = Color.Pink;
            this.d3dXfile = new D3dXfile();

            Size sz = this.pictureBox0.Size;
            this.d3dPanel.SetRender(this.d3d, this.DrawObjects, Square(sz), 
                    this.roll, this.yaw, this.pitch, this.lightCkBox.Checked);
            this.d3dPanel.BackgroundImage = global::M3dNamespace.Properties.Resources.blueBg;
            this.d3dPanel.Refresh();

            this.OffDraw();

            CameraRYP(0f, 0f, 0f);
            LightXYZ();

            this.zoomChangeEnabled = false;
            this.zoomXbox.Text = d3d.drawData.zNear.ToString("G");
            this.zoomYbox.Text = d3d.drawData.zFar.ToString("G");
            this.zoomZbox.Text = d3d.drawData.projScale.ToString("G");
            this.zoomChangeEnabled = true;
            this.UpdateZoom();
            this.lightCkBox.Checked = true;

            this.OnDraw();

            if (this.d3dXfile.IsValid() == false)
                SetStatus("Press Load button to load a model file", false);
        }

        private void pictureBox_Resize(object sender, EventArgs e)
        {
            if (this.d3dImage != null)
                 this.d3dImage.ClearImage();
            if (this.d3dPanel != null)
                this.d3dPanel.Refresh();
        }
      
        /// <summary>
        /// Draw delegate - execute all of our draw delegates.
        /// </summary>
        private void DrawObjects(D3d d3d)
        {
            // Force fill and cull mode before drawing annotation.
            // NOTE: cull mode does not work - last setting wins.
            FillMode fm = this.d3d.d3dDevice.RenderState.FillMode;
            Cull cm = this.d3d.d3dDevice.RenderState.CullMode;
            {
                this.d3d.d3dDevice.RenderState.FillMode = FillMode.Solid;
                this.d3d.d3dDevice.RenderState.CullMode = Cull.None;
                this.d3dAxis.DrawWideAxis(d3d);
                this.d3dMarkerLines.Draw(d3d);
                this.d3dNormalLines.Draw(d3d);
                this.d3dBoundingLines.Draw(d3d);
                this.d3dCenter.DrawShere(d3d, this.d3dXfile.centerV3);
            }
            this.d3d.d3dDevice.RenderState.FillMode = fm;
            this.d3d.d3dDevice.RenderState.CullMode = cm;

            // Draw xfile
            this.d3dXfile.Draw(d3d);
        }

        /// <summary>
        /// Draw (render) objects for each view (camera).
        /// </summary>
        private void Draw()
        {
            if (this.doDraw == false || d3dImage == null)
                return;

            try
            {
                Size size;
                const float f = (float)Math.PI/2.0f;

                this.Text0.Text = R3toD3(this.roll, this.yaw, this.pitch);
                if (this.d3dPanel != null)
                {
                    // On screen
                    size = Square(this.d3dPanel.Size);
                    this.d3dPanel.SetRender(this.d3d, this.DrawObjects, size,
                      this.roll, this.yaw, this.pitch, this.lightCkBox.Checked);
                }
                else
                {
                    // Off screen render.
                    size = Square(this.pictureBox0.Size);
                    this.pictureBox0.Image = d3dImage.RenderToImage(this.DrawObjects, size,
                        this.roll, this.yaw, this.pitch, this.lightCkBox.Checked);
                }

                if (d3dXfile.IsValid())
                {
                    if (!pic1Locked)
                    {
                        this.Text1.Text = R3toD3(this.roll, this.yaw + f, this.pitch);
                        this.pictureBox1.Image = d3dImage.RenderToImage(this.DrawObjects, size,
                            this.roll, this.yaw + f, this.pitch, this.lightCkBox.Checked);
                    }

                    if (!pic2Locked)
                    {
                        this.Text2.Text = R3toD3(this.roll, this.yaw, this.pitch + f);
                        this.pictureBox2.Image = d3dImage.RenderToImage(this.DrawObjects, size,
                            this.roll, this.yaw, this.pitch + f, this.lightCkBox.Checked);
                    }
                }
            }
            catch  (Exception e)
            {
                this.modelFileBox.Text = e.Message + ", Failed on: " + this.modelName;
            }
        }

        private string statusHistory = string.Empty;
        public void SetStatus(string msg, bool error)
        {
            this.Status.Text = msg;
            this.statusHistory += msg + "\n";
            this.toolTip.SetToolTip(this.Status, this.statusHistory);

            if (error)
            {
                this.Status.BackColor = Color.Red;
                this.Status.ForeColor = Color.White;
            }
            else
            {
                this.Status.BackColor = Color.White;
                this.Status.ForeColor = Color.Black;
            }
        }

        private void Status_Click(object sender, EventArgs e)
        {
            this.toolTip.Show(this.statusHistory, this.Status);
        }

        /// <summary>
        /// Load Xfile  (if .3ds, will try to convert to .x, requires conversion program).
        /// </summary>
        public void LoadModel(string xfileName)
        {
            Cursor.Current = Cursors.WaitCursor;
            Application.DoEvents();      // Dangerous

            AddMostRecent(xfileName);

            if (Path.GetExtension(xfileName).ToLower() == ext3ds)
                xfileName = Convert3dsToX(xfileName);

            this.modelName = xfileName;
            this.modelFileBox.Text = xfileName;

            this.d3dBoundingLines.ClearLines();
            this.d3dMarkerLines.ClearLines();
            this.d3dNormalLines.ClearLines(); 

            this.summaryView.Items.Clear();
            this.detailView.Items.Clear();

            if (this.d3dXfile.Load(xfileName, this.d3d))
            {
                // AutoScale
                this.d3d.drawData.projScale = 1.0f / Math.Abs(this.d3dXfile.radius * 2);
                this.d3d.drawData.zFar = Math.Max(1.0f, this.d3dXfile.radius * 2);
                this.d3d.drawData.zNear = Math.Min(-1.0f, -this.d3dXfile.radius * 2);

                this.d3dPanel.d3dBackground.BuildGrid(this.d3d, 10, 1.0f / this.d3d.drawData.projScale, this.d3d.drawData.zFar);
                this.UpdateDisplay();
                this.Text = "M3d " + Path.GetFileName(xfileName);
            }
            else if (xfileName.Length != 0)
            {
                SetStatus("Failed to load:" + xfileName + "\n" + this.d3dXfile.log.Msg, true);
            }

            Cursor.Current = Cursors.Default;
        }

        public void UpdateDisplay()
        {
            this.d3dImage.ClearImage();

            this.OffDraw();
            this.doDraw = true;
            this.Draw();

            this.zoomChangeEnabled = false;
            this.zoomXbox.Text = this.d3d.drawData.zNear.ToString("G");
            this.zoomYbox.Text = this.d3d.drawData.zFar.ToString("G");
            this.zoomZbox.Text = this.d3d.drawData.projScale.ToString("G");
            this.zoomChangeEnabled = true;
            this.UpdateZoom();

            this.OnDraw();

            this.DisplaySummary();
            this.DisplayDetail();

            this.OffDraw();
            drawNormalsCkBox_CheckedChanged(null, EventArgs.Empty);
            this.UpdateBounding();
            this.UpdateAxis();
            this.UpdateCenter();
            this.OnDraw();
            this.Draw();
        }

        /// <summary>
        /// Use MS conversion program to convert a 3DS model to an Xfile.
        /// Assumes conv3ds.exe is in execute path.
        /// </summary>
        /// <param name="in3dsFile">input 3ds filename</param>
        /// <returns>output xfilename</returns>
        string conv3dsCmd = "conv3ds.exe";
        string conv3dsOpt = "-o\"{0}\" -f -v3 {1}";
        string ext3ds = ".3ds";

        private string Convert3dsToX(string in3dsFile)
        {
            string outXFile = Path.GetFileNameWithoutExtension(in3dsFile) + "1.x";

            System.Diagnostics.Process proc = new System.Diagnostics.Process();
            try
            {
                proc.EnableRaisingEvents = false;
                proc.StartInfo.RedirectStandardOutput = false;
                proc.StartInfo.UseShellExecute = true;
                proc.StartInfo.FileName = conv3dsCmd;
                // proc.StartInfo.Arguments = "-o\"" + outXFile + "\"" + " -f -v3 " + in3dsFile;
                proc.StartInfo.Arguments = string.Format(conv3dsOpt, outXFile, in3dsFile);
                // options:
                //      -f      ; remove frame transform
                //      -r      ; reverse winding order
                //      -T      ; single Tee
                //      -s<n>   ; scale by floating point value
                //      -m      ; merge into single mesh
                //      -x      ; textural Xfile, default is binary
                //      -N      ; remove  normals
                //      -o <outfile>
                //
                proc.Start();
                proc.WaitForExit();
                proc.Dispose();

                return outXFile;
            }
            catch
            {
                return string.Empty;
            }

        }

#endregion

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.loadBtn_Click(sender, e);
        }

        private void loadBtn_Click(object sender, EventArgs e)
        {
            if (this.openModelDialog.ShowDialog() == DialogResult.OK)
            {
                string fn = this.openModelDialog.FileName;
                this.LoadModel(fn);
            }
        }

        private void closeBtn_Click(object sender, EventArgs e)
        {
            this.Close();
        }
        #endregion

        #region ===Main UI controls - Light, Camera, zoom

        /// <summary>
        /// Update light vector (update display values and re-render object)
        /// </summary>
        private void LightXYZ()
        {
            Vector3 xyz = (this.lightDir) ? this.d3d.drawData.lighting.direction : this.d3d.drawData.lighting.position;

            this.OffDraw();
            {
                this.lightX.Text = xyz.X.ToString("G");
                this.lightY.Text = xyz.Y.ToString("G");
                this.lightZ.Text = xyz.Z.ToString("G");
            }
            this.OnDraw();
            this.Draw();
        }

        /// <summary>
        /// Update camera vector (update display values and re-render object)
        /// </summary>
        private void CameraRYP(float r, float y, float p)
        {
            this.OffDraw();
            {
                this.roll = r;
                this.rollBox.Text = RtoD(this.roll).ToString("F2");

                this.yaw = y;
                this.yawBox.Text = RtoD(this.yaw).ToString("F2");

                this.pitch = p;
                this.pitchBox.Text = RtoD(this.pitch).ToString("F2");
            }
            this.OnDraw();
            this.Draw();
        }


        Vector3 mouseMove = Vector3.Empty;
        Point mousePos = Point.Empty;
        private void RollBtn_MouseMove(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left && mouseMove != Vector3.Empty)
            {
                if (mousePos != Point.Empty)
                {
                    Point delta = e.Location - new Size(mousePos);
                    float move = Math.Abs(delta.X) > Math.Abs(delta.Y) ? delta.X : delta.Y;
                    System.Diagnostics.Debug.WriteLine("move=" + move.ToString());
                    move /= 100;
                    Vector3 move3 = mouseMove * move;
                    CameraRYP(this.roll + move3.X, this.yaw + move3.Y, this.pitch + move3.Z);
                }
                mousePos = e.Location;
            }
            else
            {
                mousePos = e.Location;
                mouseMove = Vector3.Empty;
            }
        }

        private void RYP_button(object sender)
        {
            Button button = sender as Button;
            Point p = button.PointToClient(Control.MousePosition);
            int sign = (p.X < button.Width / 2) ? 1 : -1;
            Vector3 cameraMove = mouseMove * sign * 0.1f;

            CameraRYP(this.roll + cameraMove.X, this.yaw + cameraMove.Y, this.pitch + cameraMove.Z);
        }

        private void RollBtn_MouseDown(object sender, MouseEventArgs e)
        {
            mouseMove = new Vector3(1, 0, 0);
            RYP_button(sender);
        }

        private void YawBtn_MouseDown(object sender, MouseEventArgs e)
        {
            mouseMove = new Vector3(0, 1, 0);
            RYP_button(sender);
        }

        private void PitchBtn_MouseDown(object sender, MouseEventArgs e)
        {
            mouseMove = new Vector3(0, 0, 1);
            RYP_button(sender);
        }

        private void ResetBtn_Click(object sender, EventArgs e)
        {
            this.OffDraw();
            this.roll = this.yaw = this.pitch = 0.0f;
            CameraRYP(this.roll, this.yaw, this.pitch);
            this.d3d.drawData.lighting.direction = new Vector3(0, 0, 1);
            this.d3d.drawData.lighting.position = new Vector3(0, 0, 0);
            this.d3d.drawData.lighting.type = LightType.Directional;
            LightXYZ();

            float width = (this.d3dXfile.IsValid() ? this.d3dXfile.radius * 2f : 1.0f);
            this.d3d.drawData.projScale = 1.0f / width;
            this.d3d.drawData.zFar = Math.Max(1.0f, this.d3dXfile.radius * 2);
            this.d3d.drawData.zNear = Math.Min(-1.0f, -this.d3dXfile.radius * 2);

            this.d3d.drawData.translate = new Vector3(0, 0, 0);

            this.zoomChangeEnabled = false;
            this.zoomXbox.Text = this.d3d.drawData.zNear.ToString("G");
            this.zoomYbox.Text = this.d3d.drawData.zFar.ToString("G");
            this.zoomZbox.Text = this.d3d.drawData.projScale.ToString("G");
            this.zoomChangeEnabled = true;

            this.OnDraw();
            this.Draw();
        }

        private void rollBox_TextChanged(object sender, EventArgs e)
        {
            CameraRYP(DtoR(ParseToF(rollBox.Text)), this.yaw, this.pitch);
        }

        private void yawBox_TextChanged(object sender, EventArgs e)
        {
            CameraRYP(this.roll, DtoR(ParseToF(yawBox.Text)), this.pitch);
        }

        private void pitchBox_TextChanged(object sender, EventArgs e)
        {
            CameraRYP(this.roll, this.yaw, DtoR(ParseToF(pitchBox.Text)));
        }

        private void lightX_TextChanged(object sender, EventArgs e)
        {
            float n = ParseToF(lightX.Text);
            if (lightDir)
                this.d3d.drawData.lighting.direction.X = n;
            else
                this.d3d.drawData.lighting.position.X = n;

            LightXYZ();
        }

        private void lightY_TextChanged(object sender, EventArgs e)
        {
            float n = ParseToF(lightY.Text);
            if (lightDir)
                this.d3d.drawData.lighting.direction.Y = n;
            else
                this.d3d.drawData.lighting.position.Y = n;

            LightXYZ();
        }

        private void lightZ_TextChanged(object sender, EventArgs e)
        {
            float n = ParseToF(lightZ.Text);
            if (lightDir)
                this.d3d.drawData.lighting.direction.Z = n;
            else
                this.d3d.drawData.lighting.position.Z = n;

            LightXYZ();
        }

        private void lightCkBox_CheckedChanged(object sender, EventArgs e)
        {
            if (this.lightCkBox2.Checked != this.lightCkBox.Checked)
                this.lightCkBox2.Checked = this.lightCkBox.Checked;

            LightXYZ();
            this.Draw();
        }


        /// <summary>
        ///  Light tab page - check box
        /// </summary>
        private void lightCkBox2_CheckedChanged(object sender, EventArgs e)
        {
            if (this.lightCkBox.Checked != this.lightCkBox2.Checked)
                this.lightCkBox.Checked = this.lightCkBox2.Checked;
        }

        private void zoomBox_TextChanged(object sender, EventArgs e)
        {
            this.UpdateZoom();
        }

        private void UpdateZoom()
        {
            if (this.zoomChangeEnabled)
            {
                d3d.drawData.zNear = ParseToF(zoomXbox.Text);
                d3d.drawData.zFar  = ParseToF(zoomYbox.Text);
                d3d.drawData.projScale = ParseToF(zoomZbox.Text);
                if (d3d.drawData.projScale <= 0.0f)
                    d3d.drawData.projScale = 1.0f;
                this.Draw();
            }
        }
        #endregion

        #region ===Mouse movement
        private Point mouseStart = Point.Empty;

        private void pictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            int dX = (e.Location.X - this.mouseStart.X);
            int dY = (e.Location.Y - this.mouseStart.Y);

            if (e.Button == MouseButtons.Left)
            {
                if (this.mouseStart != Point.Empty)
                {
                    int d = Math.Abs(dX) > Math.Abs(dY) ? dX : dY;
                    if (Control.ModifierKeys == Keys.Control)
                    {
                        this.d3d.drawData.projScale += d / 1000.0f * this.d3d.drawData.projScale;
                        this.zoomZbox.Text = this.d3d.drawData.projScale.ToString("G");
                    }
                    else if (Control.ModifierKeys == Keys.Alt)
                    {
                        // this.d3d.drawData.zNear += d / 100.0f;
                        // this.zoomXbox.Text = this.d3d.drawData.zNear.ToString("G");
                        Vector3 translate = new Vector3(dX, -dY, 0);
                        // translate = translate * (this.d3d.drawData.projScale / 1000.0f);
                        translate = translate * (1.0f / 10.0f);
                        this.d3d.drawData.translate += translate;
                        this.Draw();
                    }
                    else if (Control.ModifierKeys == Keys.Shift)
                    {
                        this.d3d.drawData.zFar += d / 100.0f;
                        this.zoomYbox.Text = this.d3d.drawData.zFar.ToString("G");
                    }
                    else
                    {

                        this.yaw -= dX / 100.0f;
                        this.pitch += dY / 100.0f;
                        this.CameraRYP(this.roll, this.yaw, this.pitch);
                    }
                }
                this.mouseStart = e.Location;
            }
            else if (e.Button == MouseButtons.Right)
            {
                if (this.mouseStart != Point.Empty)
                {
                    if (this.lightDir)
                    {
                        d3d.drawData.lighting.direction.X -= dX / 100.0f;
                        d3d.drawData.lighting.direction.Y += dY / 100.0f;
                    }
                    else
                    {
                        d3d.drawData.lighting.position.X -= dX / 100.0f;
                        d3d.drawData.lighting.position.Y += dY / 100.0f;
                    }
                    LightXYZ();
                }
                this.mouseStart = e.Location;
            }
            else
            {
                this.mouseStart = Point.Empty;
            }
        }
        #endregion

        #region ===Keyboard hot keys
        protected override void OnKeyUp(KeyEventArgs e)
        {
            Point mousePoint = new Point(Control.MousePosition.X, Control.MousePosition.Y);
            Point d3dPoint = this.d3dView.PointToClient(mousePoint);

            ListViewItem item = this.d3dView.GetItemAt(d3dPoint.X, d3dPoint.Y);
            if (item != null)
                SetStatus(item.SubItems[0].Text, false);

            if (e.Alt)
            {
                switch (e.KeyCode)
                {
                    case Keys.Left:
                        this.d3d.drawData.translate.X -= 0.1f;
                        this.Draw();
                        return;
                    case Keys.Right:
                        this.d3d.drawData.translate.X += 0.1f;
                        this.Draw();
                        return;
                    case Keys.Up:
                        this.d3d.drawData.translate.Y += 0.1f;
                        // this.d3d.drawData.translate.Z += 0.1f;
                        this.Draw();
                        return;
                    case Keys.Down:
                        this.d3d.drawData.translate.Y -= 0.1f;
                        // this.d3d.drawData.translate.Z -= 0.1f;
                        this.Draw();
                        return;
                }
            }
            else
            {
                switch (e.KeyCode)
                {
                    case Keys.Left:
                        CameraRYP(this.roll, this.yaw - 0.1f, this.pitch);
                        return;
                    case Keys.Right:
                        CameraRYP(this.roll, this.yaw + 0.1f, this.pitch);
                        return;
                    case Keys.Up:
                        CameraRYP(this.roll, this.yaw, this.pitch - 0.1f);
                        return;
                    case Keys.Down:
                        CameraRYP(this.roll, this.yaw, this.pitch + 0.1f);
                        return;
                    case Keys.N:    // Near plane
                        if (e.Modifiers == Keys.Shift)
                            this.d3d.drawData.zNear += 0.1f;
                        else
                            this.d3d.drawData.zNear -= 0.1f;
                        this.zoomXbox.Text = this.d3d.drawData.zNear.ToString("G");
                        this.d3dImage.ClearImage();
                        // this.Draw();
                        return;
                    case Keys.F:    // Far plane
                        if (e.Modifiers == Keys.Shift)
                            this.d3d.drawData.zFar += 0.1f;
                        else
                            this.d3d.drawData.zFar -= 0.1f;
                        this.zoomYbox.Text = this.d3d.drawData.zFar.ToString("G");
                        this.d3dImage.ClearImage();
                        // this.Draw();
                        return;
                    case Keys.Oemplus:  // + = scale
                        Zoom(1.0f);
                        return;
                    case Keys.OemMinus: // - = scale
                        Zoom(-1.0f);
                        // this.Draw();
                        return;
                    case Keys.L:        // L = normals on
                        this.d3dNormalLines.BuildNormalLines(this.d3dXfile.mesh, null);
                        this.Draw();
                        return;
                    case Keys.O:        // O = normals off
                        this.d3dNormalLines.ClearLines();
                        this.Draw();
                        return;
                }
            }
            base.OnKeyUp(e);
        }
        #endregion

        #region ===Tab displays
        private void tabControl_Selected(object sender, TabControlEventArgs e)
        {
            if (e.TabPage == this.detailPage)
                DisplayDetail();
            else if (e.TabPage == this.d3dPage)
                DisplayRenderState();
        }

        public void DisplaySummary()
        {
            ListView.ListViewItemCollection summary = this.summaryView.Items;
            summary.Clear();

            if (this.d3dXfile.IsValid())
            {
                BuildMeshSummary(this.d3dXfile.mesh, summary);
            }
        }

        private void BuildMeshSummary(Mesh mesh, ListView.ListViewItemCollection summary)
        {
            if (mesh == null || summary == null)
                return;

            IndexBuffer ib = mesh.IndexBuffer;
            VertexBuffer vb = mesh.VertexBuffer;

            int na = mesh.NumberAttributes;
            int nf = mesh.NumberFaces;
            int nv = mesh.NumberVertices;

            Vector3 minV3, maxV3, centerV3;

            GraphicsStream stream = vb.Lock(0, 0, LockFlags.None);
            Geometry.ComputeBoundingBox(stream,
                mesh.NumberVertices, mesh.VertexFormat, out minV3, out maxV3);
            centerV3 = (minV3 + maxV3) * 0.5f;
            float radius = Math.Max(minV3.Length(), maxV3.Length());
            vb.Unlock();

            if (this.d3d.log.Msg.Length != 0)
            {
                string[] msgs = this.d3d.log.Msg.Split('\n');
                summary.Add("Error");
                foreach (string msg in msgs)
                {
                    summary.Add("").SubItems.Add(msg);
                }
                summary.Add(" ");
            }

            summary.Add("Vertex format").SubItems.Add(mesh.VertexFormat.ToString());
            summary.Add(" ");
            summary.Add("# Attributes").SubItems.Add(na.ToString());
            summary.Add("# Faces").SubItems.Add(nf.ToString());
            summary.Add("# Vertices").SubItems.Add(nv.ToString());
            summary.Add("Radius").SubItems.Add(radius.ToString("F2"));
            summary.Add("Min").SubItems.Add(V3ToString(minV3, "F2"));
            summary.Add("Center").SubItems.Add(V3ToString(centerV3, "F2"));
            summary.Add("Max").SubItems.Add(V3ToString(maxV3, "F2"));

            summary.Add(" ");
            string[] id = ib.Description.ToString().Split('\n');
            summary.Add("Index Buffer").SubItems.Add(id[0]);
            for (int i = 1; i < id.Length; i++)
                summary.Add(" ").SubItems.Add(id[i]);

            summary.Add(" ");
            id = vb.Description.ToString().Split('\n');
            summary.Add("Vertex Buffer").SubItems.Add(id[0]);
            for (int i = 1; i < id.Length; i++)
                summary.Add(" ").SubItems.Add(id[i]);
        }

        private void DisplayDetail()
        {
            SetStatus("Display Detail", false);
            Cursor.Current = Cursors.WaitCursor;

            ListView.ListViewItemCollection detail = this.detailView.Items;
            detail.Clear();
            Application.DoEvents();     // dangerous

            if (this.d3dXfile.IsValid())
            {
                Mesh mesh = this.d3dXfile.mesh;
                this.detailView.BeginUpdate();
                BuildMeshDetail(mesh, detail);
                this.detailView.EndUpdate();
            }

            Cursor.Current = Cursors.Default;
        }

        private void BuildMeshDetail(Mesh mesh, ListView.ListViewItemCollection detail)
        {
            if (mesh == null || detail == null)
                return;

            int nv = mesh.NumberVertices;
            int maxPoints = (int)maxKPts.Value * 1000;

            if (mesh.VertexFormat == VertexFormats.PositionNormal)
            {
                CustomVertex.PositionNormal[] vbData = (CustomVertex.PositionNormal[])
                        mesh.VertexBuffer.Lock(0, typeof(CustomVertex.PositionNormal), LockFlags.None, nv);
                ListViewItem item;

                nv = Math.Min(nv, maxPoints);    

                for (int i = 0; i < nv; i++)
                {
                    item = detail.Add(i.ToString());
                    item.SubItems.Add(ToStr(vbData[i].Position, "F3", ", "));
                    item.SubItems.Add(ToStr(vbData[i].Normal, "F3", ", "));
                    item.SubItems.Add("no-Tu");
                    item.SubItems.Add("no-Tv");
                }

                mesh.VertexBuffer.Unlock();
            }
            else
            {
                D3d.PositionNormalTexVertex[] vbData = (D3d.PositionNormalTexVertex[])
                        mesh.VertexBuffer.Lock(0, typeof(D3d.PositionNormalTexVertex), LockFlags.None, nv);
                ListViewItem item;

                nv = Math.Min(nv, maxPoints);   

                for (int i = 0; i < nv; i++)
                {
                    item = detail.Add(i.ToString());
                    item.SubItems.Add(ToStr(vbData[i].Position, "F3", ", "));
                    item.SubItems.Add(ToStr(vbData[i].Normal, "F3", ", "));
                    item.SubItems.Add(vbData[i].Tu0.ToString("F2"));
                    item.SubItems.Add(vbData[i].Tv0.ToString("F2"));
                }

                mesh.VertexBuffer.Unlock();
            }
        }

        private void detailView_ItemSelectionChanged(object sender, ListViewItemSelectionChangedEventArgs e)
        {
            if (this.detailChangedEnabled)
            {
                int[] indices = new int[this.detailView.SelectedIndices.Count];
                this.detailView.SelectedIndices.CopyTo(indices, 0);
                this.d3dMarkerLines.BuildNormalLines(d3dXfile.mesh, indices);
                this.Draw();
            }
        }

        #region Manage detailed Render State Parameters
        /// <summary>
        /// Populate ListView with current D3d Render state parameters
        /// </summary>
        private void DisplayRenderState()
        {
            d3d.RenderState.DisplayRenderState(ref this.d3dView);
        }
        /// <summary>
        /// Scrape values from ListView and pass on to D3d.
        /// </summary>
        private void UpdateRenderState()
        {
            D3dRenderState rstate = new D3dRenderState();
            d3d.RenderState = d3d.RenderState.UpdateRenderState(this.d3dView);
            this.Draw();
            this.DisplayRenderState();
        }
        /// <summary>
        /// Try and deal with key events on Listview UI
        /// </summary>
        private void d3dView_KeyUp(object sender, KeyEventArgs e)
        {
            D3dRenderState rstate = new D3dRenderState();
            if (rstate.KeyUp(this.d3dView, e))
            {
                UpdateRenderState();
            }
        }
     
        private void d3dView_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            ListViewItem item = this.d3dView.GetItemAt(e.X, e.Y);
            if (item != null)
            {
                ListViewItem.ListViewSubItem subItem = item.SubItems[1];
                Control c = sender as Control;

                if (subItem.Tag == null)
                {
                    FieldBox fieldBox = new FieldBox();
                    fieldBox.FieldText = subItem.Text;
                    fieldBox.Location = c.PointToScreen(subItem.Bounds.Location);
                    fieldBox.Width = subItem.Bounds.Width;
                    if (fieldBox.ShowDialog() == DialogResult.OK)
                    {
                        subItem.Text = fieldBox.FieldText;
                        UpdateRenderState();
                    }
                }
                else
                {
                    ComboField comboField = new ComboField();
                    comboField.ComboText = subItem.Text;
                    comboField.SetChoices(subItem.Tag as String);
                    comboField.Location = c.PointToScreen(subItem.Bounds.Location);
                    comboField.Width = subItem.Bounds.Width;
                    if (comboField.ShowDialog() == DialogResult.OK)
                    {
                        subItem.Text = comboField.ComboText;
                        UpdateRenderState();
                    }
                }
            }
        }
        #endregion

        private void UpdateNormals()
        {
            this.d3dNormalLines.color = this.normClrBtn.BackColor;
            this.d3dNormalLines.scale = (float)decimal.ToDouble(this.normLengthBox.Value) * this.d3dXfile.radius;
            this.d3dNormalLines.width = (float)decimal.ToDouble(this.normWidthBox.Value) * this.d3dXfile.radius;

            if (this.drawNormalsCkBox.Checked)
                this.d3dNormalLines.BuildNormalLines(this.d3dXfile.mesh, null);
            else
                this.d3dNormalLines.ClearLines();
            this.Draw();

            if (this.tabControl.SelectedTab == this.detailPage)
                DisplayDetail();
        }

        private void normClrBtn_Click(object sender, EventArgs e)
        {
            ColorSelDialog colorDlg = new ColorSelDialog();
            colorDlg.Color = this.normClrBtn.BackColor;
            if (colorDlg.ShowDialog(this) == DialogResult.OK)
            {
                this.normClrBtn.BackColor = colorDlg.Color;
                UpdateNormals();
            }
        }

        private void drawNormalsCkBox_CheckedChanged(object sender, EventArgs e)
        {
            UpdateNormals();  
        }

        private void normLengthBox_ValueChanged(object sender, EventArgs e)
        {
            UpdateNormals();
        }
        #endregion

        private bool pic1Locked = false;
        private void Lock1_Click(object sender, EventArgs e)
        {
            this.pic1Locked = !pic1Locked;
            this.Lock1.Image = pic1Locked ? 
                global::M3dNamespace.Properties.Resources._lock : 
                global::M3dNamespace.Properties.Resources.unlock;
        }

        private bool pic2Locked = false;
        private void Lock2_Click(object sender, EventArgs e)
        {
            this.pic2Locked = !pic2Locked;
            this.Lock2.Image = pic2Locked ?
                global::M3dNamespace.Properties.Resources._lock :
                global::M3dNamespace.Properties.Resources.unlock;
        }

        #region ===MenuBar buttons
        private ListView popupList;

        private void exportDetailBtn_Click(object sender, EventArgs e)
        {
            detailExportPopup_Click(sender, e);
        }

        private void detailExportPopup_Click(object sender, EventArgs e)
        {
            //Point p = button.PointToClient(Control.MousePosition);
            // ToolStripDropDownItem item = sender as ToolStripDropDownItem;
            if (popupList != null)
                exportList(popupList);
        }

        private void detailPopupMenu_Opening(object sender, CancelEventArgs e)
        {
            ContextMenuStrip cms = sender as ContextMenuStrip;
            popupList = cms.SourceControl as ListView;
        }

        private void keyCodesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            HelpForm helpForm = new HelpForm();
            helpForm.Title = this.Text + " Help";
            helpForm.Message = new string[17] 
            { 
                "",
                "Mouse actions: (button down and drag)",
                "  left       = rotate object",
                "  right      = move light",
                "  shift left = adjust near and far planes",
                "  ctrl left  = zoom",
                "  alt  left  = pan",
                "",
                "Key board commands:",
                "    Arrow keys        = rotate object",
                "    Alt Arrow keys   = pan object",
                "    n/N               = Adjust Near plane",
                "    f/F               = Adjust Far plane",
                "    +/-               = zoom",
                "    L/O               = normal vectors on/off",
                "    ",
                ""
            };

            helpForm.ShowDialog();

        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutBox aboutBox = new AboutBox();
            aboutBox.Show();
        }

        private void generateNormalsMenuItem_Click(object sender, EventArgs e)
        {
            if (this.d3dXfile.mesh != null)
            {
                this.d3dXfile.ComputeNormals();
                this.Draw();
                SetStatus("Normals computed on " + this.d3dXfile.mesh.NumberVertices.ToString() + " vertices", false);
            }
        }

        private void clearNormalsMemuItem_Click(object sender, EventArgs e)
        {
            this.d3dXfile.ClearNormals();
            this.Draw();
            SetStatus("Normals cleared on " + this.d3dXfile.mesh.NumberVertices.ToString() + " vertices", false);
        }

        private void reverseNormalsMenuItem_Click(object sender, EventArgs e)
        {
            int[] indices = null;

            if (this.tabControl.SelectedTab == this.detailPage && this.detailView.SelectedIndices.Count != 0)
            {
                indices = new int[this.detailView.SelectedIndices.Count];
                this.detailView.SelectedIndices.CopyTo(indices, 0);
            }

            int revCnt = this.d3dXfile.ReverseNormals(indices);
            this.Draw();
            SetStatus(revCnt.ToString() + " Normals reversed", false);
        }

        private void singleViewerToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.singleViewerToolStripMenuItem.Checked = true;
            this.threeViewersToolStripMenuItem.Checked = false;

            if (!this.pic1Locked)
                Lock1_Click(null, EventArgs.Empty);
            if (!this.pic2Locked)
                Lock2_Click(null, EventArgs.Empty);

            // int w = this.vTable.Width;
            this.vTable.ColumnStyles[0].SizeType = SizeType.Percent;
            this.vTable.ColumnStyles[1].SizeType = SizeType.Absolute;
            this.vTable.ColumnStyles[0].Width = 100f;     // 100%
            this.vTable.ColumnStyles[1].Width = 360f;     // pixels

            this.leftTable.RowStyles[0].SizeType = SizeType.Percent;
            this.leftTable.RowStyles[1].SizeType = SizeType.Absolute;
            this.leftTable.RowStyles[0].Height = 100f;    // 100%
            this.leftTable.RowStyles[1].Height = 120f;    // pixels

            this.rightTable.RowStyles[0].SizeType = SizeType.Percent;
            this.rightTable.RowStyles[1].SizeType = SizeType.Absolute;
            this.rightTable.RowStyles[0].Height = 100f;
            this.rightTable.RowStyles[1].Height = 120f;


            this.leftTable.Controls.Clear();
            this.rightTable.Controls.Clear();
            this.tabControl.Controls.Remove(this.toolBarPage);
            this.toolBarPage.Controls.Clear();
            this.summaryPage.Select();

            this.leftTable.Controls.Add(this.ulTable, 0, 0);
            this.leftTable.Controls.Add(this.toolFlow, 0, 1);
            this.rightTable.Controls.Add(this.tabControl, 0, 0);
            this.rightTable.Controls.Add(this.tblBtnPanel, 0, 1);

            this.Draw();
        }

        private void threeViewersToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.singleViewerToolStripMenuItem.Checked = false;
            this.threeViewersToolStripMenuItem.Checked = true;

            if (this.pic1Locked)
                Lock1_Click(null, EventArgs.Empty);
            if (this.pic2Locked)
                Lock2_Click(null, EventArgs.Empty);

            this.vTable.ColumnStyles[0].SizeType = SizeType.Percent;
            this.vTable.ColumnStyles[1].SizeType = SizeType.Percent;
            this.vTable.ColumnStyles[0].Width = 50f;
            this.vTable.ColumnStyles[1].Width = 50f;

            this.leftTable.RowStyles[0].SizeType = SizeType.Percent;
            this.leftTable.RowStyles[1].SizeType = SizeType.Percent;
            this.leftTable.RowStyles[0].Height = 40f;
            this.leftTable.RowStyles[1].Height = 60f;

            this.rightTable.RowStyles[0].SizeType = SizeType.Percent;
            this.rightTable.RowStyles[1].SizeType = SizeType.Percent;
            this.rightTable.RowStyles[0].Height = 40f;
            this.rightTable.RowStyles[1].Height = 60f;

            // this.picControlTable.RowStyles[0].Height = 100f;
            // this.picControlTable.RowStyles[1].Height = 34F;
            this.picControlTable.Controls.Add(this.tblBtnPanel, 0, 1);

            this.leftTable.Controls.Clear();
            this.rightTable.Controls.Clear();
            this.toolBarPage.Controls.Clear();
            
            this.leftTable.Controls.Add(this.tabControl, 0, 1);
            this.leftTable.Controls.Add(this.ulTable, 0, 0);

            this.rightTable.Controls.Add(this.picControlTable, 0, 1);
            this.rightTable.Controls.Add(this.urTable, 0, 0);

            this.toolBarPage.Controls.Add(this.toolFlow);
            if (this.tabControl.Controls.Contains(this.toolBarPage) == false)
                this.tabControl.Controls.Add(this.toolBarPage);

            this.Draw();
        }

        private void SetFillMode(FillMode fillMode)
        {
            if (this.tabControl.SelectedTab == this.renderPage)
            {
                D3dRenderState rstate = new D3dRenderState();
                rstate.UpdateRenderState(this.d3dView);
                rstate.FillMode = fillMode;
                rstate.SetRenderState(this.d3d.d3dDevice);
                this.Draw();
                this.DisplayRenderState();
            }
            else
            {
                D3dRenderState rstate = new D3dRenderState();
                rstate.GetRenderState(this.d3d.d3dDevice);
                rstate.FillMode = fillMode;
                rstate.SetRenderState(this.d3d.d3dDevice);
                this.Draw();
            }
        }

        private void solidToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetFillMode(FillMode.Solid);

            this.solidToolStripMenuItem.Checked = true;
            this.wireToolStripMenuItem.Checked = false;
            SetStatus("Solid fill mode", false);
        }

        private void wireToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SetFillMode(FillMode.WireFrame);

            this.solidToolStripMenuItem.Checked = false;
            this.wireToolStripMenuItem.Checked = true;
            SetStatus("Wireframe fill mode", false);
        }

        private void SetCullMode(Cull cull)
        {
            if (this.tabControl.SelectedTab == this.d3dPage)
            {
                D3dRenderState rstate = new D3dRenderState();
                rstate.UpdateRenderState(this.d3dView);
                rstate.CullMode = cull;
                rstate.SetRenderState(this.d3d.d3dDevice);
                this.Draw();
                this.DisplayRenderState();
                // this.d3dPage.Refresh();
            }
            else
            {
                D3dRenderState rstate = new D3dRenderState();
                rstate.GetRenderState(this.d3d.d3dDevice);
                rstate.CullMode = cull;
                rstate.SetRenderState(this.d3d.d3dDevice);
                this.Draw();
            }

            SetStatus("Cull mode set to " + cull.ToString(), false); 
        }

        private void boundingBoxToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.boundingBoxToolStripMenuItem.Checked = !this.boundingBoxToolStripMenuItem.Checked;
            this.boundingDrawCkBtn.Checked = this.boundingBoxToolStripMenuItem.Checked;
        }

        private void centerPivotToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.centerPivotToolStripMenuItem.Checked = !this.centerPivotToolStripMenuItem.Checked;
            this.centerDrawCkBtn.Checked = this.centerPivotToolStripMenuItem.Checked;
        }

        private void defaultBgViewMenuItem_Click(object sender, EventArgs e)
        {
            this.d3dPanel.BackgroundImage = global::M3dNamespace.Properties.Resources.blueBg;
            this.pictureBox0.BackgroundImage = global::M3dNamespace.Properties.Resources.blueBg;
            this.pictureBox1.BackgroundImage = global::M3dNamespace.Properties.Resources.blueBg;
            this.pictureBox2.BackgroundImage = global::M3dNamespace.Properties.Resources.blueBg;

            SetStatus("Default Background Image", false);
        }

        void OnBgColorChanged(object sender, EventArgs e)
        {
            ColorSelDialog colorDlg = sender as ColorSelDialog;
            Color bgColor = colorDlg.Color;

            this.d3dPanel.BackColor = bgColor;
            this.pictureBox0.BackColor = bgColor;
            this.pictureBox1.BackColor = bgColor;
            this.pictureBox2.BackColor = bgColor;
        }

        private void solidBgViewMenuItem_Click(object sender, EventArgs e)
        {
            this.d3dPanel.BackgroundImage = null;
            this.pictureBox0.BackgroundImage = null;
            this.pictureBox1.BackgroundImage = null;
            this.pictureBox2.BackgroundImage = null;

            ColorSelDialog ColorSelDialog = new ColorSelDialog();
            ColorSelDialog.Color = this.pictureBox0.BackColor;
            ColorSelDialog.Changed += new EventHandler(OnBgColorChanged);

            if (ColorSelDialog.ShowDialog(this) == DialogResult.OK)
                OnBgColorChanged(ColorSelDialog, EventArgs.Empty);
        }

        private void imageBgViewMenuItem_Click(object sender, EventArgs e)
        {
            if (this.bgFileDialog.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    Image bgImage = new Bitmap(this.bgFileDialog.FileName);

                    this.pictureBox0.BackgroundImage = bgImage;
                    this.pictureBox1.BackgroundImage = bgImage;
                    this.pictureBox2.BackgroundImage = bgImage;
                    this.d3dPanel.BackgroundImage = bgImage;
                    SetStatus("Background image", false);
                }
                catch { }
            }
        }

        private void saveFileMenuItem_Click(object sender, EventArgs e)
        {
            if (this.d3dXfile.IsValid())
            {
                if (this.saveFileDialog.ShowDialog() == DialogResult.OK)
                {
                    XFileFormat xfileFormat = XFileFormat.Text;
                    switch (this.saveFileDialog.FilterIndex)
                    {
                        case 0:
                            xfileFormat = XFileFormat.Text;
                            break;
                        case 1:
                            xfileFormat = XFileFormat.Binary;
                            break;
                        case 2:
                            xfileFormat = XFileFormat.Compressed;
                            break;
                        default:
                            xfileFormat = XFileFormat.Text;
                            break;
                    }

                    this.d3dXfile.Save(this.saveFileDialog.FileName, xfileFormat);
                }
            }
        }

        private void exportList(ListView listView)
        {
            if (listView == null || listView.Items.Count == 0)
                return;

            if (this.exportFileDialog.ShowDialog() == DialogResult.OK)
            {
                String filePath = this.exportFileDialog.FileName;
                TextWriter writer = new StreamWriter(filePath);

                string txtLine = string.Empty;
                foreach (ColumnHeader ch in  listView.Columns)
                {
                    if (txtLine.Length != 0)
                        txtLine += ",";
                    txtLine += ch.Text;
                }
                writer.WriteLine(txtLine);

                foreach (ListViewItem item in listView.Items)
                {
                    txtLine = string.Empty;
                    foreach (ListViewItem.ListViewSubItem subItem in item.SubItems)
                    {
                        if (txtLine.Length != 0)
                            txtLine += ",";
                        txtLine += subItem.Text.Replace(',', ';');
                    }

                    writer.WriteLine(txtLine);
                }

                writer.Close();

                SetStatus("Saved list to file:" + filePath, false);
            }
        }

        private void exportMeshFileMenuItem_Click(object sender, EventArgs e)
        {
            if (this.d3dXfile.IsValid())
                exportList(this.detailView);
        }

        private void importMeshFileMenuItem_Click(object sender, EventArgs e)
        {
            if (importFileDialog.ShowDialog() == DialogResult.OK)
            {
                MessageBox.Show("Sorry Mesh Import not yet implemented.");
                // TODO - import model mesh as csv
            }
        }

        private void snapshotImageFileMenuItem_Click(object sender, EventArgs e)
        {
            if (snapshotFileDialog.ShowDialog() == DialogResult.OK)
            {
                // TODO - allow user to control target size.
                Size size = (this.d3dPanel != null) ?
                    Square(this.d3dPanel.Size) : Square(this.pictureBox0.Size);

                try
                {
                    Bitmap image = d3dImage.RenderToImage(this.DrawObjects, size,
                                this.roll, this.yaw, this.pitch, this.lightCkBox.Checked);
                    image.Save(snapshotFileDialog.FileName);
                    SetStatus("Save snapshot to:" + snapshotFileDialog.FileName, false);
                }
                catch
                {
                    SetStatus("Failed to save snapshot to:" + snapshotFileDialog.FileName, true);
                }
            }
        }

        private void closeFileMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }


        #region === Update Axis
        private void axisToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.axisToolStripMenuItem.Checked = !this.axisToolStripMenuItem.Checked;
            if (this.axisToolStripMenuItem.Checked)
                SetStatus("Draw axis", false);
            else
                SetStatus("Clear axis", false);

            this.UpdateAxis();
        }

        private void axisLengthMenu_Click(object sender, EventArgs e)
        {
            ToolStripMenuItem menuItem = (ToolStripMenuItem)sender;
            menuItem.Checked = true;
            this.UpdateAxis();

            ToolStripDropDownMenu viewItems =
                menuItem.DropDown.OwnerItem.Owner as ToolStripDropDownMenu;
            viewItems.OwnerItem.Text = menuItem.Text;
        }

        private void axisLengthBox_ValueChanged(object sender, EventArgs e)
        {
            this.UpdateAxis();
        }

        private void axisWidthBox_ValueChanged(object sender, EventArgs e)
        {
            this.UpdateAxis();
        }

        private void axisColor_Click(object sender, EventArgs e)
        {
            Button clrBtn = sender as Button;

            ColorSelDialog colorDlg = new ColorSelDialog();
            colorDlg.Color = clrBtn.BackColor;

            if (colorDlg.ShowDialog() == DialogResult.OK)
            {
                clrBtn.BackColor = colorDlg.Color;
                this.UpdateAxis();
            }
        }

        private void UpdateAxis()
        {
            this.d3dAxis.colors = new Color[3] { this.axisXbtn.BackColor, this.axisYbtn.BackColor, this.axisZbtn.BackColor };
            float width = (float)decimal.ToDouble(this.axisWidthBox.Value) * this.d3dXfile.radius;
            float length = (float)decimal.ToDouble(this.axisLengthBox.Value) * this.d3dXfile.radius;

            if (this.axisToolStripMenuItem.Checked)
                this.d3dAxis.BuildWideLines(this.d3d, length, width);
            else
                this.d3dAxis.ClearLines();
            this.Draw();
        }
       
        #endregion

        #region === Update Center pivot point
        private void centerWidthBox_ValueChanged(object sender, EventArgs e)
        {
            this.UpdateCenter();
        }

        private void centerClrBtn_Click(object sender, EventArgs e)
        {
            Button clrBtn = sender as Button;
            ColorSelDialog colorDlg = new ColorSelDialog();
            colorDlg.Color = clrBtn.BackColor;

            if (colorDlg.ShowDialog(this) == DialogResult.OK)
            {
                clrBtn.BackColor = colorDlg.Color;
                this.UpdateCenter();
            }
        }

        private void centerDrawCkBtn_CheckedChanged(object sender, EventArgs e)
        {
            if (this.centerPivotToolStripMenuItem.Checked != this.centerDrawCkBtn.Checked)
                this.centerPivotToolStripMenuItem.Checked = this.centerDrawCkBtn.Checked;

            this.UpdateCenter();
        }

        private void UpdateCenter()
        {
            float width = (float)decimal.ToDouble(this.centerWidthBox.Value) * this.d3dXfile.radius;
            if (this.centerPivotToolStripMenuItem.Checked)
            {
                this.d3dCenter.color = this.centerClrBtn.BackColor;
                this.d3dCenter.BuildSphere(this.d3d, width);
                SetStatus("Draw Center Pivot", false);
            }
            else
            {
                this.d3dCenter.ClearSphere();
                SetStatus("Clear Center Pivot", false);
            }
            this.Draw();
        }
        #endregion

        #endregion

        #region ===Helper routines to convert and format
        /// <summary>
        /// Radians to Degree string.
        /// </summary>
        private string R3toD3(float f1, float f2, float f3)
        {
            return RtoD(f1).ToString("F2") + ", " +
                RtoD(f2).ToString("F2") + ", " +
                RtoD(f3).ToString("F2");
        }

        /// <summary>
        /// Vector3 to String
        /// </summary>
        private string V3ToString(Vector3 v3, string fmt)
        {
            return v3.X.ToString(fmt) + "," + v3.Y.ToString(fmt) + "," + v3.Z.ToString(fmt);
        }

        /// <summary>
        /// Radians to Degrees
        /// </summary>
        private double RtoD(double radians)
        {
            double degrees = radians * 180.0 / Math.PI;
            while (degrees > 180.0) degrees -= 360.0;
            while (degrees < -180.0) degrees += 360.0;
            return degrees;
        }

        /// <summary>
        ///  Degrees to Radians
        /// </summary>
        private float DtoR(float degrees)
        {
            return (float)(degrees * Math.PI / 180.0f);
        }

        /// <summary>
        /// Return square of min of input size.
        /// </summary>
        private Size Square(Size size)
        {
            int m = Math.Min(size.Width, size.Height);
            return new Size(m, m);
        }

        private string ToStr(Vector3 v3, string fmt, string sep)
        {
            return v3.X.ToString(fmt) +
                sep +
                v3.Y.ToString(fmt) +
                sep +
                v3.Z.ToString(fmt);
        }


        private float ParseToF(string inStr)
        {
            try
            {
                return float.Parse(inStr);
            }
            catch { }
            return 0f;
        }

        #endregion

        private void Zoom(float direction)
        {
            if (direction > 0)
                d3d.drawData.projScale *= 1.1f;
            else
                d3d.drawData.projScale *= 0.9f;

            zoomZbox.Text = d3d.drawData.projScale.ToString("G");
            this.Draw();
        }

        private void ZoomBtn_Click(object sender, EventArgs e)
        {
            Button button = sender as Button;
            Point p = button.PointToClient(Control.MousePosition);
            if (p.X < button.Width / 2)
            {
                Zoom(1.0f);
            }
            else
            {
                Zoom(-1.0f);
            }
        }

        private void UpdateBounding()
        {
            this.d3dBoundingLines.color = this.boundingClrBtn.BackColor;
            this.d3dBoundingLines.width = (float)Decimal.ToDouble(this.boundingWidthBox.Value);

            if (this.boundingBoxToolStripMenuItem.Checked)
            {
                this.d3dBoundingLines.BuildBoundingLines(this.d3dXfile.mesh);
                SetStatus("Draw bounding box", false);
            }
            else
            {
                this.d3dBoundingLines.ClearLines();
                SetStatus("Clear bounding box", false);
            }

            this.Draw();
        }

        private void UpdateMarker()
        {
            this.d3dMarkerLines.color = this.markClrBtn.BackColor;

            this.d3dMarkerLines.width = (float)Decimal.ToDouble(this.markWidthBtn.Value) * this.d3dXfile.radius;
            this.d3dMarkerLines.scale = (float)Decimal.ToDouble(this.markLengthBtn.Value) * this.d3dXfile.radius;

            if (this.detailView.SelectedIndices.Count > 0)
            {
                // (ListViewItemSelectionChangedEventArgs)ListViewItemSelectionChangedEventArgs.Empty);
                this.detailView_ItemSelectionChanged(null, null);

                SetStatus("Draw markers", false);
            }
            else
            {
                SetStatus("Nothing marked, use Detail tab to select vertices to mark", true);
            }
        }

        private void UpdateLight()
        {
            this.lightDir = this.lgtDirCk.Checked;
            this.lightCkBox.Text = this.lightDir ? "Light Dir" : "Light Pos";

            this.d3d.drawData.lighting.ambient = this.ambientClrBtn.BackColor;
            this.d3d.drawData.lighting.diffuse = this.diffuseClrBtn.BackColor;
            this.d3d.drawData.lighting.specular = this.specularClrBtn.BackColor;

            switch (this.lgtTypeCombo.Text)
            {
                case "Point":
                    this.d3d.drawData.lighting.type = LightType.Point;
                    break;
                case "Spot":
                    this.d3d.drawData.lighting.type = LightType.Spot;
                    break;
                case "Directional":
                    this.d3d.drawData.lighting.type = LightType.Directional;
                    break;
            }
            try
            {
                this.d3d.drawData.lighting.range = int.Parse(this.lgtRangeBox.Text);
                this.d3d.UpdateLight();
                LightXYZ();
            }
            catch { }
        }

        #region ===== Tool bar 

        private void toolIcon_Click(object sender, EventArgs e)
        {
            Button button = (Button)sender;

            switch (button.Name)
            {
                case "loadIcon":
                    this.loadBtn_Click(sender, e);
                    break;
                case "saveIcon":
                    this.saveFileMenuItem_Click(sender, e);
                    break;
                case "panel1Icon":
                    singleViewerToolStripMenuItem_Click(sender, e);
                    break;
                case "panel3Icon":
                    threeViewersToolStripMenuItem_Click(sender, e);
                    break;
                case "axisIcon":
                    axisToolStripMenuItem_Click(sender, e);
                    // this.tabControl.SelectedTab = renderPage;
                    break;
                case "boundingIcon":
                    boundingBoxToolStripMenuItem_Click(sender, e);
                    // this.tabControl.SelectedTab = renderPage;
                    break;
                case "lightIcon":
                    lightCkBox.Checked = !lightCkBox.Checked;
                    break;
                case "solidIcon":
                    solidToolStripMenuItem_Click(sender, e);
                    this.solidWireIcon.BackgroundImage = global::M3dNamespace.Properties.Resources.wireframe;
                    button.Name = "wireframeIcon";
                    break;
                case "solidWireIcon":
                case "wireframeIcon":
                    wireToolStripMenuItem_Click(sender, e);
                    this.solidWireIcon.BackgroundImage = global::M3dNamespace.Properties.Resources.solid2;
                    button.Name = "solidIcon";
                    break;
                case "paintIcon":
                    solidBgViewMenuItem_Click(sender, e);
                    break;
                case "pictureIcon":
                    imageBgViewMenuItem_Click(sender, e);
                    break;
                case "noCullIcon":
                    SetCullMode(Cull.None);
                    break;
                case "cwiseIcon":
                    SetCullMode(Cull.Clockwise);
                    break;
                case "ccwiseIcon":
                    SetCullMode(Cull.CounterClockwise);
                    break;
                case "normalIcon":
                    generateNormalsMenuItem_Click(sender, e);
                    UpdateNormals();
                    // this.tabControl.SelectedTab = renderPage;
                    break;
                case "reverseNIcon":
                    reverseNormalsMenuItem_Click(sender, e);
                    UpdateNormals();
                    // this.tabControl.SelectedTab = renderPage;
                    break;
                case "noNormalIcon":
                    clearNormalsMemuItem_Click(sender, e);
                    UpdateNormals();
                    break;                                    
            }
        }

        private void boundingClrBtn_Click(object sender, EventArgs e)
        {
            Button clrBtn = sender as Button;
            ColorSelDialog colorDlg = new ColorSelDialog();
            colorDlg.Color = clrBtn.BackColor;

            if (colorDlg.ShowDialog(this) == DialogResult.OK)
            {
                clrBtn.BackColor = colorDlg.Color;
                this.UpdateBounding();
            }
        }

        private void boundingWidthBox_ValueChanged(object sender, EventArgs e)
        {
            UpdateBounding();
        }

        private void boundingDrawCkBtn_CheckedChanged(object sender, EventArgs e)
        {
            if (this.boundingBoxToolStripMenuItem.Checked != this.boundingDrawCkBtn.Checked)
                this.boundingBoxToolStripMenuItem.Checked = this.boundingDrawCkBtn.Checked;

            UpdateBounding();
        }

        private void markClrBtn_Click(object sender, EventArgs e)
        {
            Button clrBtn = sender as Button;
            ColorSelDialog colorDlg = new ColorSelDialog();
            colorDlg.Color = clrBtn.BackColor;

            if (colorDlg.ShowDialog(this) == DialogResult.OK)
            {
                clrBtn.BackColor = colorDlg.Color;
                this.UpdateMarker();
            }
        }

        private void markLengthBtn_ValueChanged(object sender, EventArgs e)
        {
            this.UpdateMarker();
        }

        private void markWidthBtn_ValueChanged(object sender, EventArgs e)
        {
            this.UpdateMarker();
        }

 
        private void lightColor_Click(object sender, EventArgs e)
        {
            Button clrBtn = sender as Button;
            ColorSelDialog colorDlg = new ColorSelDialog();
            colorDlg.Color = clrBtn.BackColor;

            if (colorDlg.ShowDialog(this) == DialogResult.OK)
            {
                clrBtn.BackColor = colorDlg.Color;
                this.UpdateLight();
            }
        }

        private void lgtTypeCombo_SelectedIndexChanged(object sender, EventArgs e)
        {
            this.UpdateLight();
        }

        private void lgtPosCk_CheckedChanged(object sender, EventArgs e)
        {
            this.lgtDirCk.Checked = !this.lgtPosCk.Checked;
            this.UpdateLight();
        }

        private void lgtDirCk_CheckedChanged(object sender, EventArgs e)
        {
            this.lgtPosCk.Checked = !this.lgtDirCk.Checked;
            this.UpdateLight();
        }
        #endregion

        #region ==== Defailt Find/Search
        private void detailFindBox_KeyUp(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter ||
                (this.detailFindBox.Text.Length > 0 && this.detailView.Items.Count < 500))
            {
                string findStr = this.detailFindBox.Text;
                Regex regExp = new Regex("[[?*(]");
                bool isRegExp = regExp.Match(findStr).Success;
                bool select = (e.KeyCode == Keys.Enter);

                this.detailView.BeginUpdate();
                detailChangedEnabled = false;
                bool changed = SearchItems(findStr, isRegExp, select, this.detailView.Items);
                detailChangedEnabled = true;
                this.detailView.EndUpdate();
                if (changed)
                {
                    this.detailView.Invalidate();

                    // Fire selecton change callback once after we are done updating the list.
                    ListViewItemSelectionChangedEventArgs lvcE = 
                            new ListViewItemSelectionChangedEventArgs(null, -1, false);
                    detailView_ItemSelectionChanged(null, lvcE);
                }
            }
        }

        /// <summary>
        /// Search a ListView item list text for a string or regular expression.
        /// </summary>
        /// <param name="toFind">String or Expression to find</param>
        /// <param name="isRegExp">true to use regular expression</param>
        /// <param name="select">true to mark as final selection, else treat as temporary</param>
        /// <param name="lvItems">Listview Items to search</param>
        /// <returns></returns>
        private bool SearchItems(
            string toFind, 
            bool isRegExp,
            bool select,
            ListView.ListViewItemCollection lvItems)
        {
            const char sHighLight = 'h';
            int changeCnt = 0;
            Regex toFindRe = null;

            if (isRegExp)
            {
                try
                {
                    toFindRe = new Regex(toFind);
                }
                catch
                {
                    return false;
                }
            }

            foreach (ListViewItem item in lvItems)
            {
                string s = "";
                foreach (ListViewItem.ListViewSubItem subItem in item.SubItems)
                {
                    if (s.Length != 0)
                        s += ",";
                    s += subItem.Text;
                }

                bool found = (toFindRe == null) ?
                    s.Contains(toFind) : toFindRe.Match(s).Success;

                if (found)
                {
                    item.Selected = true;
                    if (select)
                        item.Tag = null;
                    else
                        item.Tag = sHighLight; // hightlight

                    if (changeCnt == 0)
                        item.Focused = true;
                    changeCnt++;
                }
                else
                {
                    if (item.Selected && item.Tag != null)
                    {
                        item.Selected = false;
                        changeCnt++;
                    }
                    item.Tag = null;
                }
            }

            return (changeCnt != 0);
        }

        #endregion

        private void addBookmarkBtn_Click(object sender, EventArgs e)
        {
            FieldBox fieldBox = new FieldBox();
            fieldBox.FieldText = "BookmarkId";

            ListViewItem item = this.bookmarkView.Items.Add(fieldBox.FieldText);
            string[] strValues = ToStrings();

            item.SubItems.Add(strValues[0]);
            item.SubItems.Add(strValues[1]);
            item.SubItems.Add(strValues[2]);
            item.SubItems.Add(strValues[3]);

            ListViewItem.ListViewSubItem subItem = item.SubItems[0];
            fieldBox.Location = this.bookmarkView.PointToScreen(subItem.Bounds.Location);
            fieldBox.Width = subItem.Bounds.Width;
            fieldBox.ShowDialog();
            item.Text = fieldBox.FieldText;
        }

        private void delBookmarkBtn_Click(object sender, EventArgs e)
        {
            this.bookmarkView.BeginUpdate();
            foreach (ListViewItem lvi in this.bookmarkView.SelectedItems)
                this.bookmarkView.Items.Remove(lvi);
            this.bookmarkView.EndUpdate();
        }

        private void clearBookmarkBtn_Click(object sender, EventArgs e)
        {
            this.bookmarkView.Items.Clear();
        }

        private void bookmarkView_DoubleClick(object sender, EventArgs e)
        {
            if (this.bookmarkView.SelectedItems.Count != 0)
            {
                ListViewItem item = this.bookmarkView.SelectedItems[0];

                string[] strValues = new String[4];
                for (int i = 0; i < strValues.Length; i++)
                    strValues[i] = item.SubItems[i + 1].Text;

                this.FromStrings(strValues);
            }
        }

        private void dragDrop(object sender, DragEventArgs e)
        {
            string fn = string.Empty;
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                string[] fns = (string[])e.Data.GetData(DataFormats.FileDrop);
                fn = fns[0];
            } 
            else if (e.Data.GetDataPresent(DataFormats.OemText))
            {
                fn = (string)e.Data.GetData(DataFormats.OemText);
            }
            else if (e.Data.GetDataPresent(DataFormats.Text))
            {
                fn = (string)e.Data.GetData(DataFormats.Text);
            }

            if (fn.Length != 0)
                this.LoadModel(fn);
        }

        private void dragEnter(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop) ||
                e.Data.GetDataPresent(DataFormats.OemText) ||
                e.Data.GetDataPresent(DataFormats.Text))
                e.Effect = DragDropEffects.Copy;
        }

        private void dragOver(object sender, DragEventArgs e)
        {
        }

        private void modelFileBox_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                this.LoadModel(modelFileBox.Text);
            }
        }

        private void registryToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string appName = Application.ProductName;

            // Preset last place RegEdit visited to our registry section
            Microsoft.Win32.RegistryKey regEditKey = Microsoft.Win32.Registry.CurrentUser.CreateSubKey(@"Software\Microsoft\Windows\CurrentVersion\Applets\Regedit");
            regEditKey.SetValue("LastKey", @"My Computer\HKEY_CURRENT_USER\Software\" + appName);
            regEditKey.Close();

            // Launch RegEdit which will default at the last place it was (our stuff).
            System.Diagnostics.Process.Start("regedit.exe");
        }

        private void dsConfigToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ConfigDialog cfgDialog = new ConfigDialog(ext3ds, conv3dsCmd, conv3dsOpt);
            if (cfgDialog.ShowDialog() == DialogResult.OK)
            {
                ext3ds = cfgDialog.Extension;
                conv3dsCmd = cfgDialog.Command;
                conv3dsOpt = cfgDialog.Options;
            }
        }

     
    }
}
