﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Runtime.InteropServices;

namespace M3dNamespace
{
    /// <summary>
    /// Author: Dennis Lang - 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// Create d3d device and support on/off screen rendering via delegates
    /// </summary>
    public class D3d : IDisposable
	{
        public bool             disposed = false;
        public Device           d3dDevice;
        public D3dDrawData      drawData;
        public Log              log;

        private D3dRenderState renderState;
        private Texture renderTex;
        private Surface renderSur;
        private RenderToSurface renderToSur;

        public delegate void DrawDelegate(D3d d3d);

        public struct PositionNormalTexVertex
        {
            public Vector3 Position;
            public Vector3 Normal;
            public float Tu0, Tv0;
            public static readonly VertexFormats FVF = VertexFormats.Position | VertexFormats.Texture1;
        }

        public D3dRenderState RenderState
        {
            get
            {
                this.renderState.GetRenderState(this.d3dDevice);
                return this.renderState;
            }
            set
            {
                this.renderState = value;
                this.renderState.SetRenderState(this.d3dDevice);
            }
        }
            
        /// <summary>
        /// Create Direct3d device.
        /// </summary>
        /// <param name="form"></param>
        public D3d(Control form)
        {
            this.log = new Log("D3d");
            this.drawData = new D3dDrawData();
            Init(form);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            if (!this.disposed)
            {
                if (disposing)
                {
                    Clear();
                    d3dDevice.Dispose();
                }

                // Call appropriate methods to release any other resources
                // such as handles and raw pointers.

                this.disposed = true;
            }
        }

        /// <summary>
        /// Clear various textures and surfaces
        /// </summary>
        public void Clear()
        {
            if (this.renderTex != null)
            {
                this.renderSur.Dispose();
                this.renderToSur.Dispose();
                this.renderTex.Dispose();
                this.renderTex = null;
            }
        }

		/// <summary>
		/// Creates and initialize the Direct3D device.
		/// </summary>
		// private void Init(System.Windows.Forms.Form form)
		private void Init(Control form)
		{
			//
			// Do we support hardware vertex processing? If so, use it. 
			// If not, downgrade to software.
			//
			Caps caps = Manager.GetDeviceCaps(
                Manager.Adapters.Default.Adapter, DeviceType.Hardware );
			CreateFlags flags;

			if( caps.DeviceCaps.SupportsHardwareTransformAndLight )
				flags = CreateFlags.HardwareVertexProcessing;
			else
				flags = CreateFlags.SoftwareVertexProcessing;

            DisplayMode dspMode = Manager.Adapters[0].CurrentDisplayMode;

			//
			// Everything checks out - create a simple, windowed device.
			//
			PresentParameters d3dpp = new PresentParameters();
            d3dpp.BackBufferFormat       = dspMode.Format;
			d3dpp.SwapEffect             = SwapEffect.Discard;
			d3dpp.Windowed               = true;
			d3dpp.EnableAutoDepthStencil = true;
            d3dpp.AutoDepthStencilFormat = DepthFormat.D16;

            d3dDevice = new Device(0, DeviceType.Hardware, form, flags, d3dpp);

            renderState = new D3dRenderState();
            renderState.GetRenderState(d3dDevice);
            renderState.ZBufferEnable = true;
            renderState.CullMode = Cull.None;

			// Register an event-handler for DeviceReset and call it to continue
			// our setup.
			d3dDevice.DeviceReset += new System.EventHandler(this.OnResetDevice);
			OnResetDevice(d3dDevice, null);
		}

		/// <summary>
		/// This event-handler is a good place to create and initialize any 
		/// Direct3D related objects, which may become invalid during a 
		/// device reset.
		/// </summary>
		public void OnResetDevice(object sender, EventArgs e)
		{
            Device device = (Device)sender;

            renderState.SetRenderState(device);
            device.RenderState.Lighting = true;
            // device.RenderState.DiffuseMaterialSource = ColorSource.Material;

            UpdateLight();
        }

        public void UpdateLight()
        {
            // Setup lighting
            d3dDevice.Lights[0].Enabled = this.drawData.lighting.enabled;
            d3dDevice.Lights[0].Type = this.drawData.lighting.type;
            d3dDevice.Lights[0].Ambient = this.drawData.lighting.ambient;
            d3dDevice.Lights[0].Diffuse = this.drawData.lighting.diffuse;
            d3dDevice.Lights[0].Range = this.drawData.lighting.range;
        }
       
        public Vector2 ToScreen(Vector3 worldV3)
        {
            Matrix finalMx = d3dDevice.GetTransform(TransformType.Projection);
            Vector4 screenV4 = Vector3.Transform(worldV3, finalMx);

            return new Vector2(screenV4.X, screenV4.Y);
        }

        public Matrix ProjMatrix
        {
            get { return d3dDevice.GetTransform(TransformType.Projection); }
        }

        public Matrix ViewMatrix
        {
            get { return d3dDevice.GetTransform(TransformType.View); }
        }

        public Matrix WorldMatrix
        {
            get { return d3dDevice.GetTransform(TransformType.World);}
        }


		/// <summary>
		/// Render to screen, requires a window.
		/// </summary>
        public void Render(DrawDelegate drawDelegate, Size size,
            float roll, float yaw, float pitch, bool lightOn)
        {
            float invScale = 1.0f / this.drawData.projScale;
#if false
            Matrix matView = Matrix.LookAtLH(
                new Vector3(0.0f, 0.0f, -cameraZPos),   // cameraPosition
                new Vector3(0.0f, 0.0f, 0.0f),          // cameraTangent
                new Vector3(0.0f, 1.0f, 0.0f));         // cameraUpVector
            Matrix matProj = Matrix.PerspectiveFovLH(invScale/cameraZPos, 1.0f, zNear, zFar);
#else
            Matrix matProj = Matrix.OrthoLH(invScale, invScale, this.drawData.zNear, this.drawData.zFar);
            Matrix matView = Matrix.Identity;
#endif
            Matrix matWorldR = Matrix.Identity;
            matWorldR = Matrix.RotationYawPitchRoll(yaw, pitch, roll);
            Matrix matWorldT = Matrix.Identity;
            matWorldT.Translate(this.drawData.translate);
            Matrix matWorld = matWorldR * matWorldT;

            ClearFlags clearFlags = ClearFlags.Target;
            if (d3dDevice.RenderState.ZBufferEnable)
                clearFlags |= ClearFlags.ZBuffer;

            float aspect = (float)size.Height / size.Width;
            Vector3 aspect3 = new Vector3(aspect, 1.0f, 1.0f);
            matView.Scale(aspect3);

            d3dDevice.BeginScene();
            {
                d3dDevice.Clear(clearFlags, Color.Transparent, 1.0f, 0);
                d3dDevice.SetTransform(TransformType.Projection, matProj);
                d3dDevice.SetTransform(TransformType.World, matWorld);
                d3dDevice.SetTransform(TransformType.View, matView);

                if (lightOn)
                {
                    // Setup the lights
                    d3dDevice.Lights[0].Direction = this.drawData.lighting.direction;
                    d3dDevice.Lights[0].Position = this.drawData.lighting.position;
                    d3dDevice.Lights[0].Falloff = 1f;
                    d3dDevice.Lights[0].InnerConeAngle = Geometry.DegreeToRadian(10.0f);
                    d3dDevice.Lights[0].OuterConeAngle = Geometry.DegreeToRadian(40.0f);
                    d3dDevice.Lights[0].Attenuation0 = 1f; 
                    d3dDevice.Lights[0].Enabled = true;
                    d3dDevice.RenderState.Lighting = true;
                }
                else
                {
                    d3dDevice.Lights[0].Enabled = false;
                    d3dDevice.RenderState.Lighting = false;
                }

                if (drawDelegate != null)
                {
                    drawDelegate(this);
                }
            }
            d3dDevice.EndScene();
            d3dDevice.Present();
        }

        /// <summary>
        ///  Render off screen to a surface, calling drawDelegate.
        ///  Off screen means you do not use a window to render and do not see the results on the screen.
        /// </summary>
        /// <param name="drawDelegate">delegate to draw mesh or lines</param>
        /// <param name="size">Draw surface size</param>
        /// <param name="roll">camera roll</param>
        /// <param name="yaw">camera yaw</param>
        /// <param name="pitch">camera pictch</param>
        /// <param name="lightOn">lighting on or off</param>
        /// <returns></returns>
        public Surface RenderOffScreenTexture(DrawDelegate drawDelegate, Size size, 
            float roll, float yaw, float pitch, bool lightOn)
        {
            float invScale = 1.0f / this.drawData.projScale;
#if false
            // Setup Perspective projection
            Matrix matView = Matrix.LookAtLH(
                new Vector3(0.0f, 0.0f, -cameraZPos),   // cameraPosition
                new Vector3(0.0f, 0.0f, 0.0f),          // cameraTangent
                new Vector3(0.0f, 1.0f, 0.0f));         // cameraUpVector
            Matrix matProj = Matrix.PerspectiveFovLH(invScale/cameraZPos, 1.0f, zNear, zFar);
#else
            // Setup Ortho projection
            Matrix matProj = Matrix.OrthoLH(invScale, invScale, this.drawData.zNear, this.drawData.zFar);
            Matrix matView = Matrix.Identity;
#endif
            Matrix matWorldR = Matrix.Identity;
            matWorldR = Matrix.RotationYawPitchRoll(yaw, pitch, roll);
            Matrix matWorldT = Matrix.Identity;
            matWorldT.Translate(this.drawData.translate);
            Matrix matWorld = matWorldR * matWorldT;


            if (renderTex == null || renderTex.Disposed)
            {
                renderTex = new Texture(d3dDevice, size.Width, size.Height, 1,
                    Usage.RenderTarget, Format.A8R8G8B8, Pool.Default);

                renderSur = renderTex.GetSurfaceLevel(0);
                SurfaceDescription surDesc = renderSur.Description;

                // d3dDevice.CreateDepthStencilSurface(size.Width, size.Height, DepthFormat.D16, 
                //     MultiSampleType.None, 0, true);

                renderToSur = new RenderToSurface(d3dDevice,
                        surDesc.Width, surDesc.Height,
                        surDesc.Format, true, DepthFormat.D16);
            }

            ClearFlags clearFlags = ClearFlags.Target;
            if (d3dDevice.RenderState.ZBufferEnable)
                clearFlags |= ClearFlags.ZBuffer;

            renderToSur.BeginScene(renderSur);
            {
                d3dDevice.Clear(clearFlags, Color.Transparent, 1.0f, 0);
                d3dDevice.SetTransform(TransformType.Projection, matProj);
                d3dDevice.SetTransform(TransformType.World, matWorld);
                d3dDevice.SetTransform(TransformType.View, matView);

                if (lightOn)
                {
                    // Setup the lights
                    d3dDevice.Lights[0].Enabled = true;
                    d3dDevice.Lights[0].Direction = this.drawData.lighting.direction;
                    d3dDevice.Lights[0].Position = this.drawData.lighting.position;
                    d3dDevice.RenderState.Lighting = true;
                }
                else
                {
                    d3dDevice.Lights[0].Enabled = false;
                    d3dDevice.RenderState.Lighting = false;
                }

                if (drawDelegate != null)
                {
                    drawDelegate(this);
                }
            }

            renderToSur.EndScene(Filter.Linear);
            return renderSur;
        }
    }
}
