﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using System.Windows;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Runtime.InteropServices;

namespace M3dNamespace
{
    /// Author: Dennis Lang - 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// Draw colorized wide lines. 
    /// 
    class D3dColorLineList
    {
        public float width = 1.0f;
        public float length = 1.0f;
        public Color[] colors = new Color[3] { Color.Red, Color.Green, Color.Blue };

        private CustomVertex.PositionColored[] linePcList;
        private Mesh wideLine;

        public void BuildWideLines(D3d d3d, float Length, float Width)
        {
            int stacks = 4;
            int slicesAround = 16;
            this.length = Length;
            this.width = Width;
            wideLine = Mesh.Cylinder(d3d.d3dDevice, Width / 2.0f, Width / 2.0f, length, slicesAround, stacks);
            wideLine.ComputeNormals();

            {
                int nv = wideLine.NumberVertices;

                CustomVertex.PositionNormal[] vbData = (CustomVertex.PositionNormal[])
                    wideLine.VertexBuffer.Lock(0, typeof(CustomVertex.PositionNormal), LockFlags.None, nv);

                float zEnd = length / 2.0f;
                Vector3 zTip = new Vector3(0f, 0f, zEnd);
                zEnd *= 0.9f;

                for (int i = 0; i < nv; i++)
                {
                    if (vbData[i].Position.Z >= zEnd)
                        vbData[i].Position = zTip;
                    if (vbData[i].Position.Z == 0.0f)
                        vbData[i].Position = vbData[i].Position * 1.3f;
                }

                wideLine.VertexBuffer.Unlock();
            }

            wideLine.ComputeNormals();
        }

        /// <summary>
        ///  Build lines pointing along major coordinate axis.
        /// </summary>
        public void BuildAxis(float axisSize)
        {
            this.linePcList = new CustomVertex.PositionColored[3 * 2];

            Vector3 origin = new Vector3(0, 0, 0);
            linePcList[0].Position = origin;
            linePcList[0].Color = colors[0].ToArgb();
            linePcList[1].Position = origin +  new Vector3(axisSize, 0, 0);
            linePcList[1].Color = colors[0].ToArgb();

            linePcList[2].Position = origin;
            linePcList[2].Color = colors[1].ToArgb();
            linePcList[3].Position = origin +  new Vector3(0, axisSize, 0);
            linePcList[3].Color = colors[1].ToArgb();

            linePcList[4].Position = origin;
            linePcList[4].Color = colors[2].ToArgb();
            linePcList[5].Position = origin +  new Vector3(0, 0, axisSize);
            linePcList[5].Color = colors[2].ToArgb();
        }

        public void ClearLines()
        {
            this.linePcList = null;
            this.wideLine = null;
        }

        public void Draw(D3d d3d)
        {
            if (this.linePcList != null)
            {
                d3d.d3dDevice.SetTexture(0, null);      // clear texture else its color is used.

                // Save color, set new color, draw, restore original color
                Color c = d3d.d3dDevice.RenderState.Ambient;
                {
                    d3d.d3dDevice.RenderState.Ambient = Color.White;
                    d3d.d3dDevice.RenderState.PointSize = this.width;
                    d3d.d3dDevice.VertexFormat = CustomVertex.PositionColored.Format;
                    d3d.d3dDevice.DrawUserPrimitives(PrimitiveType.LineList, linePcList.Length / 2, linePcList);
                }
                // Restore color
                d3d.d3dDevice.RenderState.Ambient = c;
            }
        }

        /// <summary>
        /// Draw arrows pointing along major coordinate axis. 
        /// </summary>
        public void DrawWideAxis(D3d d3d)
        {
            if (this.wideLine != null)
            {
                d3d.d3dDevice.SetTexture(0, null);      // clear texture else its color is used.
                Material[] m = new Material[2];
                m[0] = d3d.d3dDevice.Material;
                m[1].Ambient = Color.Black;
                m[1].Diffuse = Color.Black;
                d3d.d3dDevice.Material = m[1];

                // Save color, set new color, draw, restore original color
                Color c = d3d.d3dDevice.RenderState.Ambient;
                Matrix w = d3d.d3dDevice.Transform.World;
                {
                    d3d.d3dDevice.RenderState.Ambient = Color.White;

                    Matrix rotationMatrix = Matrix.Identity;
                    Matrix translationMatrix = Matrix.Identity;

                    // Cylinder mesh length along z-axis, centered on xyz

                    // Draw x-axis RED
                    translationMatrix.Translate(length/2.0f, 0.0f, 0.0f);
                    rotationMatrix.RotateY(Geometry.DegreeToRadian(90.0f));
                    d3d.d3dDevice.Transform.World = rotationMatrix * translationMatrix * w;
                    m[1].Diffuse = this.colors[0];
                    d3d.d3dDevice.Material = m[1];

                    this.wideLine.DrawSubset(0);

                    // Draw y-axis GREEN
                    rotationMatrix.RotateX(Geometry.DegreeToRadian(-90.0f));
                    translationMatrix.Translate(0.0f, length / 2.0f, 0.0f);
                    d3d.d3dDevice.Transform.World = rotationMatrix * translationMatrix * w;
                    m[1].Diffuse = this.colors[1];
                    d3d.d3dDevice.Material = m[1];
                    this.wideLine.DrawSubset(0);

                    // Draw z-axis BLUE
                    translationMatrix.Translate(0.0f, 0.0f, length / 2.0f);
                    d3d.d3dDevice.Transform.World = translationMatrix * w;
                    m[1].Diffuse = this.colors[2];
                    d3d.d3dDevice.Material = m[1];
                    this.wideLine.DrawSubset(0);
                }
                // Restore color
                d3d.d3dDevice.RenderState.Ambient = c;
                d3d.d3dDevice.Transform.World = w;
                d3d.d3dDevice.Material = m[0];
            }
        }
    }
}
