﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using System.Windows;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Runtime.InteropServices;

namespace M3dNamespace
{
    /// Author: Dennis Lang - 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// Create a grid mesh and optionally attach an image.
    /// Use to build a background to the viewer. 
    /// 
    public class D3dGrid
    {
        public Color color = Color.Blue;

        private int[] indices;
        private IndexBuffer ib;
        private VertexBuffer vb;
        private Texture tex;
        private Size size;
        private float scale = 1.0f;
        private float depth = 0.0f;
        private CustomVertex.PositionNormalTextured[] verts;
        private Bitmap image;

        /// <summary>
        /// Set/Get background image (texture)
        /// </summary>
        public Image BackgroundImage
        {
            get { return image;}
            set
            {
                this.image = (Bitmap)value;
                this.tex = null;
            }
        }
        public Color BackColor
        {
            get { return this.color; }
            set
            {
                this.color = value;
                this.tex = null;
            }
        }

        /// <summary>
        ///  Depth(z-coordinate) grid drawn
        /// </summary>
        public float Depth
        {
            get { return this.depth; }
        }

        /// <summary>
        ///  Draw scale.
        /// </summary>
        public float Scale
        {
            get { return this.scale; }
        }

        /// <summary>
        /// Build grid 
        /// </summary>
        /// <param name="d3d"></param>
        /// <param name="numCellsAcross"></param>
        /// <param name="inScale"></param>
        /// <param name="inDepth"></param>
        public void BuildGrid(D3d d3d, int numCellsAcross, float inScale, float inDepth)
        {
            this.depth = inDepth;
            this.scale = inScale;
            this.size = new Size(numCellsAcross, numCellsAcross);

            int gridWidth = size.Width;
            int gridHeight = size.Height;
            int gridWidthM1 = gridWidth - 1;
            int gridHeightM1 = gridHeight - 1;

            vb = new VertexBuffer(typeof(CustomVertex.PositionNormalTextured), gridWidth * gridHeight,
                    d3d.d3dDevice, Usage.Dynamic | Usage.WriteOnly, CustomVertex.PositionNormalTextured.Format, 
                    Pool.Default);
            ib = new IndexBuffer(typeof(int), (gridWidthM1) * (gridHeightM1) * 6,
                    d3d.d3dDevice, Usage.WriteOnly, Pool.Default);

            indices = new int[(gridWidthM1) * (gridHeightM1) * 6];
            for (int x = 0; x < gridWidthM1; x++)
            {
                for (int y = 0; y < gridHeightM1; y++)
                {
                    int off = (x + y * gridWidthM1) * 6;

                    indices[off + 0] = (x + 1) + (y + 1) * gridWidth;
                    indices[off + 1] = (x + 1) + (y + 0) * gridWidth;
                    indices[off + 2] = (x + 0) + (y + 0) * gridWidth;

                    indices[off + 3] = (x + 1) + (y + 1) * gridWidth;
                    indices[off + 4] = (x + 0) + (y + 0) * gridWidth;
                    indices[off + 5] = (x + 0) + (y + 1) * gridWidth;
                }
            }

            float u, v;
            float pX, pY, pZ;
            float nX, nY, nZ;
            pZ = depth;        
            nZ = 1f;

            verts = new CustomVertex.PositionNormalTextured[gridWidth * gridHeight];

            for (int x = gridWidthM1; x >= 0; x--)
            {
                u = x & 1;
                nX = pX = (float)x / gridWidthM1 * scale;

                for (int y = gridHeightM1; y >= 0; y--)
                {
                    v = y & 1;
                    nY = pY = (float)y / gridHeightM1 * scale;

                    verts[x + y * gridWidth] = new CustomVertex.PositionNormalTextured(
                        pX, pY, pZ, nX, nY, nZ , u, v);
                }
            }

            ib.Created += new EventHandler(this.OnIndexBufferCreate);
            OnIndexBufferCreate(ib, EventArgs.Empty);
            vb.Created += new EventHandler(this.OnVertexBufferCreate);
            OnVertexBufferCreate(vb, EventArgs.Empty);
        }

        private void OnIndexBufferCreate(object sender, EventArgs e)
        {
            IndexBuffer buffer = (IndexBuffer)sender;
            buffer.SetData(indices, 0, LockFlags.None);
        }
        private void OnVertexBufferCreate(object sender, EventArgs e)
        {
            VertexBuffer buffer = (VertexBuffer)sender;
            buffer.SetData(verts, 0, LockFlags.None);
        }

        public void SetTexture(D3d d3d, Bitmap texImage)
        {
            d3d.d3dDevice.SetTexture(0, null);
            if (texImage != null && texImage.Width > 0)
            {
                d3d.d3dDevice.SetTexture(0, null);
                this.tex = null;
                this.image = texImage;
                this.color = Color.White;
            }
        }
       
        public void Draw(D3d d3d)
        {
            if (this.vb != null)
            {
                Material[] m = new Material[2];
                m[0] = d3d.d3dDevice.Material;          // save material
                m[1].Ambient = Color.White;
                m[1].Diffuse = this.color;
                d3d.d3dDevice.Material = m[1];

                int gridWidth = size.Width;
                int gridHeight = size.Height;

                Matrix w = d3d.d3dDevice.Transform.World;
                Color c = d3d.d3dDevice.RenderState.Ambient;
                {
                    d3d.d3dDevice.RenderState.Ambient = this.color;

                    // Matrix translationMatrix = Matrix.Identity;
                    // d3d.d3dDevice.Transform.World = translationMatrix * w;

                    if (tex == null && image != null)
                        this.tex = Texture.FromBitmap(d3d.d3dDevice, this.image, Usage.None, Pool.Managed);

                    d3d.d3dDevice.SetTexture(0, tex);

                    d3d.d3dDevice.VertexFormat = CustomVertex.PositionNormalTextured.Format;
                    d3d.d3dDevice.SetStreamSource(0, vb, 0); 
                    d3d.d3dDevice.Indices = ib;
                    d3d.d3dDevice.Transform.World = Matrix.Translation(-scale/2f, -scale/2f, 0f);
                    d3d.d3dDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, gridWidth * gridHeight, 0,
                          indices.Length / 3);
                }
                d3d.d3dDevice.RenderState.Ambient = c;  // restore color
                d3d.d3dDevice.Material = m[0];          // restore material
                d3d.d3dDevice.Transform.World = w;      // restore transform
            }
        }
    }
}