﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using System.Windows;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Runtime.InteropServices;

namespace M3dNamespace
{
    /// Author: Dennis Lang - 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// <summary>
    /// Render Direct3D object to an image.
    /// </summary>
    public class D3dImage
    {
        private Surface surLocal = null;
        private Bitmap finalImage = null;
        private D3d d3d;

        public Log log;

        public D3dImage(D3d d3dRef)
        {
            log = new Log("D3dImage");
            d3d = d3dRef;
        }

        public void ClearImage()
        {
            this.log.Clear();
            this.surLocal = null;
            this.finalImage = null;

            this.d3d.Clear();
        }

        private Bitmap FillImageFromSurface(
            Bitmap dstBitmap, 
            Surface inSurface, 
            System.Drawing.Imaging.PixelFormat format)
        {
            // Populate the destination image, by copying the surface pixels 
            // into the bitmap image buffer.
            SurfaceDescription inDesc = inSurface.Description;

            if (dstBitmap == null)
                dstBitmap = new Bitmap(inDesc.Width, inDesc.Height, format);

            Rectangle rect = new Rectangle(0, 0, inDesc.Width, inDesc.Height);

            int pitch;
            System.Drawing.Imaging.BitmapData dstData =
                dstBitmap.LockBits(rect, System.Drawing.Imaging.ImageLockMode.WriteOnly, format);

            Int32[] srcInt32 = (Int32[])inSurface.LockRectangle(typeof(Int32),
                LockFlags.ReadOnly, out pitch, inDesc.Width * inDesc.Height);

            if (dstData.Stride != pitch)
            {
                this.log.Msg += "We got a pitch/stride problem ";
            }

            try
            {
                unsafe
                {
                    Int32* dstInt32 = (Int32*)dstData.Scan0;
                    Marshal.Copy(srcInt32, 0, dstData.Scan0, inDesc.Width * inDesc.Height);
                }
            }
            catch
            {
                // error on line h, column w.
                this.log.Msg += "Failure during image copy on row:" + inDesc.Height.ToString()
                    + " colum:" + inDesc.Width.ToString();
            }

            inSurface.UnlockRectangle();
            dstBitmap.UnlockBits(dstData);
            return dstBitmap;
        }

        /// <summary>
        /// Render Direct3D to an image.
        /// </summary>
        /// <param name="drawDelegate">delegate to call during Draw pass</param>
        /// <param name="size">Target image size</param>
        /// <param name="roll">Camera roll</param>
        /// <param name="yaw"></param>
        /// <param name="pitch"></param>
        /// <param name="lightOn">Lighting state</param>
        /// <returns></returns>
        public Bitmap RenderToImage(D3d.DrawDelegate drawDelegate, Size size,
            float roll, float yaw, float pitch, bool lightOn)
        {
            // Render to image takes three steps:
            //  1. Render model using D3d to surface (call delegates)
            //  2. Copy remote surface to local surface (wish I could skip this)
            //  3. Copy local surface to image

            // 1. Render d3d object offscreen
            Surface surRender = this.d3d.RenderOffScreenTexture(drawDelegate, size,
                roll, yaw, pitch, lightOn);

            // Create local surface
            if (surLocal == null || surLocal.Disposed)
            {
                Bitmap bitmap = new Bitmap(size.Width, size.Height,
                    System.Drawing.Imaging.PixelFormat.Format32bppArgb);
                surLocal = new Surface(this.d3d.d3dDevice, bitmap, Pool.SystemMemory);
            }

            // 2. Copy render surface to local surface
            this.d3d.d3dDevice.GetRenderTargetData(surRender, surLocal);

            // 3. Copy surface in to image.
            finalImage = FillImageFromSurface(null, surLocal,
                System.Drawing.Imaging.PixelFormat.Format32bppArgb);

            surRender.Dispose();

            return finalImage;
        }
    }
}
