﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using System.Windows;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Runtime.InteropServices;

namespace M3dNamespace
{
    /// Author: Dennis Lang - 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// Draw one or more lines.  To see line in color enable lighting.
    /// 
    class D3dLineList
    {
        public float scale = 0.3f;
        public float width = 3.0f;  // line thickness
        public Color color = Color.Red;

        private CustomVertex.PositionOnly[] lineList = null;
        private Vector3[] vector3List;
        private Line line;

        private void BuildVectorList()
        {
            vector3List = new Vector3[lineList.Length];
            for (int i = 0; i < lineList.Length; i++)
                vector3List[i] = lineList[i].Position;
        }


        /// <summary>
        ///  Build lines pointing along normal vectors
        /// </summary>
        /// <param name="mesh">Input mesh containing normals</param>
        /// <param name="markerIndexes">Array of offsets to normals to process</param>
        public void BuildNormalLines(Mesh mesh, int[] markerIndexes)
        {
            if (mesh != null)
            {
                int nv = mesh.NumberVertices;

                if (mesh.VertexFormat == VertexFormats.PositionNormal)
                {
                    CustomVertex.PositionNormal[] vbData = (CustomVertex.PositionNormal[])
                        mesh.VertexBuffer.Lock(0, typeof(CustomVertex.PositionNormal), LockFlags.ReadOnly, nv);

                    int n = 0;
                    if (markerIndexes == null)
                    {
                        this.lineList = new CustomVertex.PositionOnly[nv * 2];
                        for (int i = 0; i < nv; i++)
                        {
                            this.lineList[n].Position = vbData[i].Position;
                            this.lineList[n + 1].Position = vbData[i].Position + vbData[i].Normal * this.scale;
                            n += 2;
                        }
                    }
                    else
                    {
                        this.lineList = new CustomVertex.PositionOnly[markerIndexes.Length * 2];
                        foreach (int i in markerIndexes)
                        {
                            this.lineList[n].Position = vbData[i].Position;
                            this.lineList[n + 1].Position = vbData[i].Position + vbData[i].Normal * this.scale;
                            n += 2;
                        }
                    }
                    mesh.VertexBuffer.Unlock();
                }
                else
                {
                    D3d.PositionNormalTexVertex[] vbData = (D3d.PositionNormalTexVertex[])
                        mesh.VertexBuffer.Lock(0, typeof(D3d.PositionNormalTexVertex), LockFlags.ReadOnly, nv);

                    int n = 0;
                    if (markerIndexes == null)
                    {
                        this.lineList = new CustomVertex.PositionOnly[nv * 2];
                        for (int i = 0; i < nv; i++)
                        {
                            this.lineList[n].Position = vbData[i].Position;
                            this.lineList[n + 1].Position = vbData[i].Position + vbData[i].Normal * this.scale;
                            n += 2;
                        }
                    }
                    else
                    {
                        this.lineList = new CustomVertex.PositionOnly[markerIndexes.Length * 2];
                        foreach (int i in markerIndexes)
                        {
                            this.lineList[n].Position = vbData[i].Position;
                            this.lineList[n + 1].Position = vbData[i].Position + vbData[i].Normal * this.scale;
                            n += 2;
                        }
                    }
                    mesh.VertexBuffer.Unlock();
                }

                BuildVectorList();
            }
        }

        public void BuildBoundingLines(Mesh mesh)
        {
            if (mesh == null)
                return;

            Vector3 minV3, maxV3;

            VertexBuffer vertices = mesh.VertexBuffer;
            GraphicsStream stream = vertices.Lock(0, 0, LockFlags.None);
            Geometry.ComputeBoundingBox(stream,
                mesh.NumberVertices, mesh.VertexFormat, out minV3, out maxV3);
            vertices.Unlock();

            this.lineList = new CustomVertex.PositionOnly[12 * 2];

            int n = 0;
            // Connect top and bottom plan
            this.lineList[n++].Position = minV3;
            this.lineList[n++].Position = new Vector3(minV3.X, minV3.Y, maxV3.Z);

            this.lineList[n++].Position = new Vector3(minV3.X, maxV3.Y, minV3.Z);
            this.lineList[n++].Position = new Vector3(minV3.X, maxV3.Y, maxV3.Z);

            this.lineList[n++].Position = new Vector3(maxV3.X, minV3.Y, minV3.Z);
            this.lineList[n++].Position = new Vector3(maxV3.X, minV3.Y, maxV3.Z);

            this.lineList[n++].Position = new Vector3(maxV3.X, maxV3.Y, minV3.Z);
            this.lineList[n++].Position = new Vector3(maxV3.X, maxV3.Y, maxV3.Z);

            // Ring bottom plan
            this.lineList[n++].Position = new Vector3(minV3.X, minV3.Y, minV3.Z);
            this.lineList[n++].Position = new Vector3(minV3.X, maxV3.Y, minV3.Z);

            this.lineList[n++].Position = new Vector3(minV3.X, maxV3.Y, minV3.Z);
            this.lineList[n++].Position = new Vector3(maxV3.X, maxV3.Y, minV3.Z);

            this.lineList[n++].Position = new Vector3(maxV3.X, maxV3.Y, minV3.Z);
            this.lineList[n++].Position = new Vector3(maxV3.X, minV3.Y, minV3.Z);

            this.lineList[n++].Position = new Vector3(maxV3.X, minV3.Y, minV3.Z);
            this.lineList[n++].Position = new Vector3(minV3.X, minV3.Y, minV3.Z);

            // Ring top plan
            this.lineList[n++].Position = new Vector3(minV3.X, minV3.Y, maxV3.Z);
            this.lineList[n++].Position = new Vector3(minV3.X, maxV3.Y, maxV3.Z);

            this.lineList[n++].Position = new Vector3(minV3.X, maxV3.Y, maxV3.Z);
            this.lineList[n++].Position = new Vector3(maxV3.X, maxV3.Y, maxV3.Z);

            this.lineList[n++].Position = new Vector3(maxV3.X, maxV3.Y, maxV3.Z);
            this.lineList[n++].Position = new Vector3(maxV3.X, minV3.Y, maxV3.Z);

            this.lineList[n++].Position = new Vector3(maxV3.X, minV3.Y, maxV3.Z);
            this.lineList[n++].Position = new Vector3(minV3.X, minV3.Y, maxV3.Z);

            BuildVectorList();
        }

        public void ClearLines()
        {
            this.lineList = null;
            this.vector3List = null;
        }

        public void Draw(D3d d3d)
        {
            if (this.lineList != null && this.lineList.Length >= 2)
            {
                d3d.d3dDevice.SetTexture(0, null);      // clear texture else its color is used.

                // Save color, set new color, draw, restore original color
                Color c = d3d.d3dDevice.RenderState.Ambient;
                d3d.d3dDevice.RenderState.Ambient = this.color;

                if (width <= 1.0f)
                {
                    d3d.d3dDevice.RenderState.PointSize = this.width;
                    d3d.d3dDevice.VertexFormat = CustomVertex.PositionOnly.Format;
                    d3d.d3dDevice.DrawUserPrimitives(PrimitiveType.LineList, lineList.Length / 2, lineList);
                }
                else
                {
                    // Draw wide lines
                    if (this.line == null)
                        this.line = new Line(d3d.d3dDevice);
                    this.line.Width = width;

                    Matrix wpv = d3d.WorldMatrix * d3d.ProjMatrix * d3d.ViewMatrix;
                    Vector3[] lineSeg = new Vector3[2];
                    for (int nLine = 0; nLine < vector3List.Length; nLine += 2)
                    {
                        lineSeg[0] = vector3List[nLine + 0];
                        lineSeg[1] = vector3List[nLine + 1];
                        this.line.DrawTransform(lineSeg, wpv, this.color);
                    }
                }

                // Restore color
                d3d.d3dDevice.RenderState.Ambient = c;
            }
        }
    }
}
