﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace M3dNamespace
{
    /// Author: Dennis Lang - 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// <summary>
    ///  Panel which paints its contents by using D3d
    ///  See SetRender and delegate Draw method.
    /// </summary>
    public class D3DPanel : Panel
    {
        private D3d d3d;

        public D3DPanel()
        {
            // Prevent D3d flashing.
            this.SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque, true);
        }

        private D3d.DrawDelegate drawDelegate;
        private Size drawSize;
        private float roll, yaw, pitch;
        private bool lightOn;
        public D3dGrid d3dBackground;

        public void Clear()
        {
            d3dBackground = null;
        }

        /// <summary>
        /// Set background color (causes background image to be ignored).
        /// </summary>
        public override Color BackColor
        {
            get 
            { 
                if  (this.d3dBackground == null)
                    return Color.White;
                else
                    return this.d3dBackground.BackColor; 
            }
            set 
            {
                if (this.d3dBackground == null)
                    this.BackColor = value;
                else
                    this.d3dBackground.BackColor = value;
            }
        }

        /// <summary>
        /// Set background image (causes background color to be ignored).
        /// </summary>
        public override Image BackgroundImage
        {
            get
            {
                return this.d3dBackground.BackgroundImage;
            }
            set
            {
                this.d3dBackground.color = Color.White;
                this.d3dBackground.BackgroundImage = value;
            }
        }

        /// <summary>
        /// Setup D3d device and delegate draw method.
        /// Specify view rotation and whether lights are on.
        /// </summary>
        public void SetRender(D3d inD3d, D3d.DrawDelegate inDrawDelegate, Size inSize,
            float inRoll, float inYaw, float inPitch, bool inLightOn)
        {
            this.d3d = inD3d;
            this.drawDelegate = inDrawDelegate;
            this.drawSize = inSize;
            this.roll = inRoll;
            this.yaw = inYaw;
            this.pitch = inPitch;
            this.lightOn = inLightOn;

            if (this.d3dBackground == null)
                this.d3dBackground = new D3dGrid();

            float scale = 1.0f / this.d3d.drawData.projScale;
            if (this.d3dBackground.Depth != this.d3d.drawData.zFar ||
                this.d3dBackground.Scale != scale)
            {
                this.d3dBackground.BuildGrid(d3d, 10,  scale, this.d3d.drawData.zFar);
            }

            this.Refresh();
        }

        /// <summary>
        ///  Private delegate to draw background correctly (solid and no culling).
        /// </summary>
        /// <param name="d3d"></param>
        private void BackgroundDelegate(D3d d3d)
        {
            FillMode fm = this.d3d.d3dDevice.RenderState.FillMode;
            Cull     cm = this.d3d.d3dDevice.RenderState.CullMode;
            {
                this.d3d.d3dDevice.RenderState.FillMode = FillMode.Solid;
                this.d3d.d3dDevice.RenderState.CullMode = Cull.None;
                this.d3dBackground.Draw(d3d);
            }
            this.d3d.d3dDevice.RenderState.CullMode = cm;
            this.d3d.d3dDevice.RenderState.FillMode = fm;
        }

        /// <summary>
        /// Draw d3d objects, assumes z-buffer is enabled and draws delegates first
        /// with background grid drawn last.
        /// </summary>
        /// <param name="d3d"></param>
        private void DrawObjects(D3d d3d)
        {
            // Assuming depth buffer is enabled, draw object first
            this.drawDelegate(d3d);

            // Draw background last
            BackgroundDelegate(d3d);
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            if (this.drawSize != null)
                this.d3d.Render(DrawObjects, DisplayRectangle.Size, roll, yaw, pitch, lightOn);
        }
    }
}
