﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Drawing;
using System.Windows;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Runtime.InteropServices;

namespace M3dNamespace
{
    /// Author: Dennis Lang - 2009
    /// http://home.comcast.net/~lang.dennis/
    /// 
    /// <summary>
    /// Load XFile and provide drawDelegate to render
    /// </summary>
    class D3dXfile
    {
        // Result of load
        public VertexBuffer vertices = null;
        public Mesh mesh = null;
        public ExtendedMaterial[] materialarray = null;
        public Material[] meshmaterials = null;
        public Texture[] meshtextures = null;

        // Status information.
        public Log log = new Log("D3dXfile");
        public Vector3 minV3, maxV3, centerV3;
        public float radius;


        /// <summary>
        /// Return true if contains a valid mesh.
        /// </summary>
        public bool IsValid()
        {
            return (mesh != null && mesh.NumberVertices > 0);
        }

        /// <summary>
        /// Save mesh to disk file.
        /// </summary>
        /// <param name="filename">destination x filename</param>
        /// <param name="xfileFormat">file format</param>
        public void Save(string filename, XFileFormat xfileFormat)
        {
            VertexBuffer vertices = mesh.VertexBuffer;
            GraphicsStream gstream = vertices.Lock(0, 0, LockFlags.None);
            vertices.Unlock();

            this.mesh.Save(filename, gstream, this.materialarray, xfileFormat);
        }

        /// <summary>
        /// Load Xfile into memory.
        /// Throw exception if invalid.
        /// </summary>
        /// <param name="filename">Xfile name</param>
        /// <param name="d3d">D3d device</param>
        /// <returns>true if all okay, else inspect Log for info</returns>
        public bool Load(string filename, D3d d3d)
        {
            log.Clear();

            // Quick check for bad file
            if (filename == null || filename.Length == 0 || filename.Trim().Length == 0)
                return false;

            try
            {
                // Better check for bad file.
                FileInfo fileInfo;
                fileInfo = new FileInfo(filename);
            
                if (fileInfo.Exists)
                {
                    // Try and load Xfile
                    mesh = Mesh.FromFile(filename, MeshFlags.Managed, d3d.d3dDevice, out materialarray);

                    if ((materialarray != null) && (materialarray.Length > 0))
                    {
                        meshmaterials = new Material[materialarray.Length];
                        meshtextures = new Texture[materialarray.Length];

                        for (int i = 0; i < materialarray.Length; i++)
                        {
                            meshmaterials[i] = materialarray[i].Material3D;
                            meshmaterials[i].Ambient = meshmaterials[i].Diffuse;

                            if ((materialarray[i].TextureFilename != null) && (materialarray[i].TextureFilename != string.Empty))
                            {
                                string texFilename = materialarray[i].TextureFilename;
                                FileInfo finfo = new FileInfo(texFilename);
                                if (finfo.Exists == false && Path.GetDirectoryName(texFilename).Length == 0)
                                {
                                    texFilename = Path.GetDirectoryName(filename) + Path.DirectorySeparatorChar +
                                        Path.GetFileName(texFilename);
                                    finfo = new FileInfo(texFilename);
                                }
                                if (finfo.Exists)
                                {
                                    try
                                    {
                                        meshtextures[i] = TextureLoader.FromFile(d3d.d3dDevice, texFilename);
                                    }
                                    catch
                                    {
                                        log.Msg += "Exception trying to load material image:" + texFilename + "\n";
                                    }
                                }
                                else
                                {
                                    log.Msg += "Missing material image:" + texFilename + "\n";
                                }
                            }
                        }
                    }

                    mesh = mesh.Clone(mesh.Options.Value, CustomVertex.PositionNormalTextured.Format, d3d.d3dDevice);

                    vertices = mesh.VertexBuffer;
                    GraphicsStream stream = vertices.Lock(0, 0, LockFlags.None);

                    Geometry.ComputeBoundingBox(stream,
                        mesh.NumberVertices, mesh.VertexFormat, out minV3, out maxV3);
                    centerV3 = (minV3 + maxV3) * 0.5f;
                    radius = Math.Max(minV3.Length(), maxV3.Length());

                    // float scaling = 1.0f;
                    // radius = Geometry.ComputeBoundingSphere(stream,
                    //     mesh.NumberVertices, mesh.VertexFormat, out centerV3) * scaling;
                    vertices.Unlock();
                    return true;
                }
            }
            catch (Exception e)
            {
                log.Msg += "Invalid file " + e.ToString();
                return false;
            }

            return false;
        }

        /// <summary>
        /// Draw delegate to render Xfile
        /// </summary>
        public  void Draw(D3d d3d)
        {
            if (meshmaterials != null)
            {
                for (int i = 0; i < meshmaterials.Length; i++)
                {
                    d3d.d3dDevice.Material = meshmaterials[i];
                    d3d.d3dDevice.SetTexture(0, meshtextures[i]);
                    mesh.DrawSubset(i);
                }
            }
            else
            {
                d3d.d3dDevice.SetTexture(0, null);
                if (mesh != null)
                    mesh.DrawSubset(0);
            }
        }

        /// <summary>
        /// Generate normals by averaging vertices
        /// </summary>
        public void ComputeNormals()
        {
            if (mesh != null)
                mesh.ComputeNormals();
        }

        public void ClearNormals()
        {
            if (mesh != null)
            {
                int nv = mesh.NumberVertices;
                Vector3 zero = new Vector3(0f, 0f, 0f);

                if (mesh.VertexFormat == VertexFormats.PositionNormal)
                {
                    CustomVertex.PositionNormal[] vbData = (CustomVertex.PositionNormal[])
                        mesh.VertexBuffer.Lock(0, typeof(CustomVertex.PositionNormal), LockFlags.None, nv);

                    for (int i = 0; i < nv; i++)
                        vbData[i].Normal = zero;

                    mesh.VertexBuffer.Unlock();
                }
                else
                {
                    D3d.PositionNormalTexVertex[] vbData = (D3d.PositionNormalTexVertex[])
                        mesh.VertexBuffer.Lock(0, typeof(D3d.PositionNormalTexVertex), LockFlags.None, nv);

                    for (int i = 0; i < nv; i++)
                        vbData[i].Normal = zero;

                    mesh.VertexBuffer.Unlock();
                }
            }
        }

        /// <summary>
        /// Reverse normals.
        /// </summary>
        /// <param name="indexes">List of normal indexes to reverse</param>
        public int ReverseNormals(int[] indexes)
        {
            if (mesh != null)
            {
                int nv = mesh.NumberVertices;

                D3d.PositionNormalTexVertex[] vbData = (D3d.PositionNormalTexVertex[])
                    mesh.VertexBuffer.Lock(0, typeof(D3d.PositionNormalTexVertex), LockFlags.None, nv);

                if (indexes == null)
                {
                    for (int i = 0; i < nv; i++)
                        vbData[i].Normal *= -1.0f;
                }
                else
                {
                    int n = 0;
                    foreach (int i in indexes)
                    {
                        vbData[i].Normal *= -1.0f;
                        n += 2;
                    }
                }
                mesh.VertexBuffer.Unlock();
                return (indexes == null) ? nv : indexes.Count();
            }
            return 0;
        }
    }
}
